from bokeh.palettes import viridis,magma
from bokeh.models import ColumnDataSource, Label, LabelSet,CustomJS
from bokeh.layouts import column,row,gridplot
from bokeh.plotting import figure, show
from bokeh.models import LinearColorMapper, LogTicker,BasicTicker, ColorBar,LogColorMapper,Legend
from bokeh.palettes import magma as colfun1
from bokeh.palettes import Colorblind8
from bokeh.palettes import viridis as colfun2
from bokeh.palettes import gray as colfun3
from bokeh.palettes import Colorblind8
import bokeh.palettes as colpals
from bokeh.io import output_notebook 
import astropy.units as u
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter # used to display axes labels as 0.1 instead of 10^-1 power notation
from matplotlib.ticker import FormatStrFormatter
from statistics import mean
from scipy.signal import savgol_filter
from matplotlib.legend_handler import HandlerPathCollection
import matplotlib
import pandas as pd
import sys
from .root_functions import calculate_k0, find_N_mon_and_r_mon

from . import justdoit as pyeddy

def pt(out, with_condensation=True,return_condensation=False, **kwargs):
    """
    Plot PT profiles with condensation curves. 
    Thick lines in this pt plot signify those gases that 
    were turned on in the run, NOT those that were recommended. 
    
    Parameters
    ----------
    out : dict 
        Dictionary output from pyeddy run 
    with_condensation : bool 
        Plots condensation curves of gases. Also plots those that were turned on in the 
        calculation with line_width=5. All others are set to 1. 
    return_condensation : bool 
        If true, it returns list of condenation temps for each gas, pressure grid, 
        and a list of the gas names
    **kwargs : kwargs 
        Kwargs for bokeh.figure() 

    Returns
    -------
    if return_condensation: fig, cond temperature curves, ,cond pressure grid , name of all available gases
    else: fig
    """
    kwargs['height'] = kwargs.get('plot_height',kwargs.get('height',300))
    kwargs['width'] = kwargs.get('plot_width', kwargs.get('width',600))
    if 'plot_width' in kwargs.keys() : kwargs.pop('plot_width')
    if 'plot_height' in kwargs.keys() : kwargs.pop('plot_height')
    kwargs['x_axis_label'] = kwargs.get('x_axis_label','Temperature (K)')
    kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure (bars)')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')
    

    ngas = len(out['condensibles'])
    temperature = out['temperature']
    pressure = out['pressure']
    condensibles = out['condensibles']
    mh, mmw = out['scalar_inputs']['mh'], out['scalar_inputs']['mmw']

    kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure), np.min(pressure)])

    fig = figure(**kwargs)

    if with_condensation:
        all_gases = pyeddy.available()
        cond_ts = []
        recommend = []
        line_width = []

        for gas_name in all_gases: #case sensitive names
            #grab p,t from eddysed
            cond_p,t = pyeddy.condensation_t(gas_name, mh, mmw)
            cond_ts +=[t]
            if gas_name in condensibles:
                line_width += [5]
            else: 
                line_width += [1]

        cols = magma(len(all_gases))
        for i in range(len(all_gases)):
            fig.line(cond_ts[i],cond_p, legend_label=all_gases[i],color=cols[i],line_width=line_width[i] )

    fig.line(temperature,pressure, legend_label='User',color='black',line_width=5,line_dash='dashed')

    plot_format(fig)
    if return_condensation: 
        return fig, cond_ts,cond_p,all_gases
    else : 
        return fig


def radii(out,gas=None,at_pressure = 1e-3, compare=False, legend=None, 
        p1w=300, p1h=300, p2w=300, p2h=300, color_indx=0):
    """
    Plots the particle radii profile along with the distribution, at a certain pressure. 

    Parameters
    ----------
    out : dict 
        Dictionary output from pyeddy run 
    pressure : float,optional 
        Pressure level to plot full distribution (bars)
    
    """
    if type(out)==dict:
        out=[out]

    lines = ['solid','dashed','dashdot']
    #lines = ['solid']*3
    if compare:
        lines = ['solid']*len(out)
        legend_it = []
    for j in range(len(out)):
        #compute initial distributions   
        r_g = out[j]['mean_particle_r']
        pressure = out[j]['pressure']

        nl = find_nearest_1d(pressure,at_pressure)

        rmin = out[j]['scalar_inputs']['rmin']
        rmax = out[j]['scalar_inputs']['rmax']
        nrad = out[j]['scalar_inputs']['nrad']
        log_radii = out[j]['scalar_inputs']['log_radii']
        if(log_radii>0):
            logspace=True # convert scalar variable into boolean for get_r_grid
        else:
            logspace=False # convert scalar variable into boolean for get_r_grid
        sig = out[j]['scalar_inputs']['sig']
        ndz = out[j]['column_density']
        #determine which condensed
        which_condensed = [False]*ndz.shape[1]

        # only consider specified gas
        if gas is not None:
            indx = out[j]['condensibles'].index(gas)
            ndz_gas = np.unique(ndz[:,indx])
            ndz_gas = ndz_gas[ndz_gas>0]
            if len(ndz_gas)>0 : 
                which_condensed[indx] = True
            else:
                print(condensate + " did not condense. Choose another condensate.")
                import sys; sys.exit()

        # consider all gases
        else:
            for i in range(ndz.shape[1]):
                ndz_gas = np.unique(ndz[:,i])
                ndz_gas = ndz_gas[ndz_gas>0]
                if len(ndz_gas)>0 : which_condensed[i] = True


        #take only those that condensed 
        N = ndz[:, which_condensed]
        r_g = r_g[:,which_condensed]
        gas_name = list(np.array(out[j]['condensibles'])[which_condensed])
        r, bin_min, bin_max, dr = pyeddy.get_r_grid(r_min = rmin, r_max=rmax, n_radii = nrad, log_space=logspace)

        # different colours for different dicts
        if (gas is not None) or compare:
            length = len(out)
        # different colours for different gases
        else:
            length = len(gas_name) 

        color = magma(length)
        color = Colorblind8[color_indx:color_indx+length]

        #initial radii profiles
        df_r_g = {i:r_g[:, gas_name.index(i)] for i in gas_name}
        df_r_g['pressure'] = pressure


        dndr = {}
        for i in gas_name:
            dndr[i]= N[nl,gas_name.index(i)]/r/np.sqrt(2*np.pi)*np.log(sig)*np.exp(
                        - np.log(r/(r_g[nl,gas_name.index(i)]*1e-4))**2/(2*np.log(sig)**2)) #1e-4 is microns to cm
        dndr['r'] = r*1e4 #convert to microns

        if j==0:
            p1 = figure(width=p1w, height=p1h, title="Select Pressure Level", 
                x_axis_type='log',y_axis_type='log',y_axis_label='Pressure (bars)',x_axis_label='Mean Particle Radius (um)',
                y_range=[np.max(pressure),np.min(pressure)])

            p2 = figure(width=p2w, height=p2h, title="Particle Distribution at %.0e bars" %at_pressure,
                         x_axis_type='log', y_axis_type='log', y_axis_label='dn/dr (cm-3)',
                         x_axis_label='Particle Radius (um)', 
                         y_range=[np.max([np.min(dndr[i]), 1e-50]), np.max(dndr[i])])

        #add to r_g for that plot 
        df_r_g['average'] = [pressure[nl]]*len(pressure)
        df_r_g['horizontal'] = np.linspace(np.amin(r_g[r_g>0]), np.amax(r_g) , len(pressure))
            

        s1 = ColumnDataSource(data=dict(df_r_g))
        s2 = ColumnDataSource(data=dict(dndr))
        for i in gas_name:
            if gas is not None or compare: indx = j
            else: indx = gas_name.index(i)
            f = p1.line(i, 'pressure', source=s1, alpha=1,color=color[np.mod(indx, len(color))] ,line_width=4,legend_label=i,line_dash=lines[j])
            p2.line('r', i, source=s2,color=color[np.mod(indx, len(color))] ,line_width=4,line_dash=lines[j])

            if compare:
                legend_it.append((legend[j], [f]))

    # r_g = out['droplet_eff_r']
    # pressure = out['pressure']

    # nl = find_nearest_1d(pressure,at_pressure)

    # rmin = out['scalar_inputs']['rmin']
    # nrad = out['scalar_inputs']['nrad']
    # sig = out['scalar_inputs']['sig']
    # ndz = out['column_density']
    # #determine which condensed
    # which_condensed = [False]*ndz.shape[1]
    # for i in range(ndz.shape[1]):
    #     ndz_gas = np.unique(ndz[:,i])
    #     ndz_gas = ndz_gas[ndz_gas>0]
    #     if len(ndz_gas)>0 : which_condensed[i] = True

    # color = magma(len(which_condensed))

    # #take only those that condensed 
    # N = ndz[:, which_condensed]
    # r_g = r_g[:,which_condensed]
    # gas_name = list(np.array(out['condensibles'])[which_condensed])
    # r, rup, dr = pyeddy.get_r_grid(r_min = rmin, n_radii = nrad)

    # #initial radii profiles
    # df_r_g = {i:r_g[:, gas_name.index(i)] for i in gas_name}
    # df_r_g['pressure'] = pressure


    # #add to r_g for that plot 
    # df_r_g['average'] = [pressure[nl]]*len(pressure)
    # df_r_g['horizontal'] = np.linspace(np.amin(r_g[r_g>0]), np.amax(r_g) , len(pressure))

    p1.line('horizontal', 'average', source=s1, color='black',line_width=3,line_dash='dashed')
    p1.legend.location = 'bottom_left'
    #plot_format(p1)
    #plot_format(p2)
    if compare:
        legend = Legend(items=legend_it, location=(0, 0))
        legend.click_policy="mute"
        p1.add_layout(legend, 'right')   
    return row(p1, p2), dndr

def opd_by_gas(out, gas = None, color = magma, compare=False, legend=None, **kwargs):
    """
    Optical depth for conservative geometric scatteres separated by gas.
    E.g. [Fig 7 in Morley+2012](https://arxiv.org/pdf/1206.4313.pdf)  
    
    Parameters
    ----------
    out : dict 
        Dictionary output from pyeddy run
    color : method
        Method from bokeh.palletes. e.g. bokeh.palletes.virids, bokeh.palletes.magma
    **kwargs : kwargs 
        Kwargs for bokeh.figure() 
    """

    kwargs['height'] = kwargs.get('plot_height',kwargs.get('height',300))
    kwargs['width'] = kwargs.get('plot_width', kwargs.get('width',400))
    if 'plot_width' in kwargs.keys() : kwargs.pop('plot_width')
    if 'plot_height' in kwargs.keys() : kwargs.pop('plot_height')
    kwargs['x_axis_label'] = kwargs.get('x_axis_label','Column Optical Depth')
    kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure (bars)')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')
    kwargs['x_axis_type'] = kwargs.get('x_axis_type','log')    

    if type(out)==dict:
        out=[out]

    condensibles = out[0]['condensibles']
    if gas is not None: 
        indx = condensibles.index(gas)
        length = len(out)
    else:
        length = len(condensibles)
    ngas = len(condensibles)
    col = color(length)
    col = Colorblind8[:length]
    lines = ['solid','dashed','dotdash','dashdot']

    legend_it = []
    for j in range(len(out)):
        pressure = out[j]['pressure']
        opd_by_gas = out[j]['opd_by_gas']

        if j == 0:
            kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure), np.min(pressure)])
            kwargs['x_range'] = kwargs.get('x_range',[np.max([1e-6, np.min(opd_by_gas*0.9)]), 
                                                np.max(opd_by_gas*1.1)])
            fig = figure(**kwargs)

        if compare:
            if gas is not None:
                f = fig.line(opd_by_gas[:,indx], pressure, line_width=4, legend_label = condensibles[indx], color=Colorblind8[np.mod(j, len(Colorblind8))])
            else:
                for i in range(ngas):
                    f = fig.line(opd_by_gas[:,i], pressure,line_width=4, legend_label = condensibles[i], color=Colorblind8[np.mod(j, len(Colorblind8))], line_dash=lines[i])
            legend_it.append((legend[j], [f]))
        else:
            if gas is not None:
                fig.line(opd_by_gas[:,indx], pressure, line_width=4, legend_label = condensibles[indx], color=col[j], line_dash=lines[j])
            else:
                for i in range(ngas):
                    fig.line(opd_by_gas[:,i], pressure,line_width=4, legend_label = condensibles[i], color=col[i], line_dash=lines[j])

    if compare:
        legend = Legend(items=legend_it, location=(0, 0))
        legend.click_policy="mute"
        fig.add_layout(legend, 'right')   
    #plot_format(fig)
    return fig

def condensate_mmr(out, gas=None, compare=False, legend=None, **kwargs):
    """
    Condensate mean mass mixing ratio
    
    Parameters
    ----------
    out : dict 
        Dictionary output from pyeddy run
    color : method
        Method from bokeh.palletes. e.g. bokeh.palletes.virids, bokeh.palletes.magma
    **kwargs : kwargs 
        Kwargs for bokeh.figure() 
    """
    kwargs['height'] = kwargs.get('plot_height',kwargs.get('height',300))
    kwargs['width'] = kwargs.get('plot_width', kwargs.get('width',400))
    if 'plot_width' in kwargs.keys() : kwargs.pop('plot_width')
    if 'plot_height' in kwargs.keys() : kwargs.pop('plot_height')
    kwargs['x_axis_label'] = kwargs.get('x_axis_label','Condensate MMR')
    kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure (bars)')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')
    kwargs['x_axis_type'] = kwargs.get('x_axis_type','log')    

    if type(out)==dict:
        out=[out]

    condensibles = out[0]['condensibles']
    if gas is not None :
        indx = condensibles.index(gas)
        length = len(out)
    else:
        length = len(condensibles)
    ngas = len(condensibles)
    col = magma(length)
    lines = ['solid','dashed','dotdash','dashdot']

    legend_it = []
    for j in range(len(out)):
        pressure = out[j]['pressure']
        cond_mmr = out[j]['condensate_mmr']

        if j == 0:
            kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure), np.min(pressure)])
            kwargs['x_range'] = kwargs.get('x_range',[np.max([1e-9, np.min(cond_mmr*0.9)]), 
                                                    np.max(cond_mmr*1.1)])
            fig = figure(**kwargs)

        if compare:
            if gas is not None:
                f = fig.line(cond_mmr[:,i], pressure, line_width=4, legend_label = condensibles[indx], color=Colorblind8[np.mod(j, len(Colorblind8))])
            else:
                for i in range(ngas):
                    label = condensibles[i] + ' ' + legend[j]
                    f = fig.line(cond_mmr[:,i], pressure, line_width=4, legend_label = condensibles[i], color=Colorblind8[np.mod(j, len(Colorblind8))], line_dash=lines[i])
            legend_it.append((legend[j], [f]))

        else:
            if gas is not None:
                fig.line(cond_mmr[:,indx], pressure, line_width=4, legend_label = condensibles[indx], color=Colorblind8[np.mod(j, len(Colorblind8))], line_dash=lines[j])
            else:
                for i in range(ngas):
                    fig.line(cond_mmr[:,i], pressure, line_width=4, legend_label = condensibles[i], color=col[i], line_dash=lines[j])

    if compare:
        legend = Legend(items=legend_it, location=(0, 0))
        legend.click_policy="mute"
        fig.add_layout(legend, 'right')   

    plot_format(fig)
    return fig

def all_optics(out):
    """
    Maps of the wavelength dependent single scattering albedo 
    and cloud opacity and asymmetry parameter as a function of altitude. 

    Parameters
    ----------
    out : dict 
        Dictionary output from pyeddy run 

    Returns
    -------
    Three bokeh plots with the single scattering, optical depth, and assymetry maps
    """
    #get into DataFrame format
    dat01 = pyeddy.picaso_format(out['opd_per_layer'],out['single_scattering'],
                      out['asymmetry'])

    nwno=len(out['wave'])
    nlayer=len(out['pressure'])
    pressure=out['pressure']

    pressure_label = 'Pressure (Bars)'

    wavelength_label = 'Wavelength (um)'
    wavelength = out['wave']

    cols = colfun1(200)
    color_mapper = LinearColorMapper(palette=cols, low=0, high=1)

    #PLOT W0
    scat01 = np.flip(np.reshape(dat01['w0'].values,(nlayer,nwno)),0)
    xr, yr = scat01.shape
    f01a = figure(x_range=[0, yr], y_range=[0,xr],
                           x_axis_label=wavelength_label, y_axis_label=pressure_label,
                           title="Single Scattering Albedo",
                          width=300, height=300)


    f01a.image(image=[scat01],  color_mapper=color_mapper, x=0,y=0,dh=xr,dw =yr )

    color_bar = ColorBar(color_mapper=color_mapper, #ticker=LogTicker(),
                       label_standoff=12, border_line_color=None, location=(0,0))

    f01a.add_layout(color_bar, 'left')


    #PLOT OPD
    scat01 = np.flip(np.reshape(dat01['opd'].values,(nlayer,nwno)),0)

    xr, yr = scat01.shape
    cols = colfun2(200)[::-1]
    color_mapper = LogColorMapper(palette=cols, low=1e-3, high=10)


    f01 = figure(x_range=[0, yr], y_range=[0,xr],
                           x_axis_label=wavelength_label, y_axis_label=pressure_label,
                           title="Cloud Optical Depth Per Layer",
                          width=320, height=300)

    f01.image(image=[scat01],  color_mapper=color_mapper, x=0,y=0,dh=xr,dw =yr )

    color_bar = ColorBar(color_mapper=color_mapper, ticker=LogTicker(),
                       label_standoff=12, border_line_color=None, location=(0,0))
    f01.add_layout(color_bar, 'left')

    #PLOT G0
    scat01 = np.flip(np.reshape(dat01['g0'].values,(nlayer,nwno)),0)

    xr, yr = scat01.shape
    cols = colfun3(200)[::-1]
    color_mapper = LinearColorMapper(palette=cols, low=0, high=1)


    f01b = figure(x_range=[0, yr], y_range=[0,xr],
                           x_axis_label=wavelength_label, y_axis_label=pressure_label,
                           title="Asymmetry Parameter",
                          width=300, height=300)

    f01b.image(image=[scat01],  color_mapper=color_mapper, x=0,y=0,dh=xr,dw =yr )

    color_bar = ColorBar(color_mapper=color_mapper, ticker=BasicTicker(),
                       label_standoff=12, border_line_color=None, location=(0,0))
    f01b.add_layout(color_bar, 'left')

    #CHANGE X AND Y AXIS TO BE PHYSICAL UNITS 
    #indexes for pressure plot 
    if (pressure is not None):
        pressure = ["{:.1E}".format(i) for i in pressure[::-1]] #flip since we are also flipping matrices
        npres = len(pressure)
        ipres = np.array(range(npres))
        #set how many we actually want to put on the figure 
        #hard code ten on each.. 
        ipres = ipres[::int(npres/10)]
        pressure = pressure[::int(npres/10)]
        #create dictionary for tick marks 
        ptick = {int(i):j for i,j in zip(ipres,pressure)}
        for i in [f01a, f01, f01b]:
            i.yaxis.ticker = ipres
            i.yaxis.major_label_overrides = ptick
    if (wavelength is not None):
        wave = ["{:.2F}".format(i) for i in wavelength]
        nwave = len(wave)
        iwave = np.array(range(nwave))
        iwave = iwave[::int(nwave/10)]
        wave = wave[::int(nwave/10)]
        wtick = {int(i):j for i,j in zip(iwave,wave)}
        for i in [f01a, f01, f01b]:
            i.xaxis.ticker = iwave
            i.xaxis.major_label_overrides = wtick       

    return gridplot([[f01a, f01,f01b]])

def all_optics_1d(out, wave_range, return_output = False,legend=None,
    colors = colpals.Colorblind8, **kwargs):
    """
    Plots 1d profiles of optical depth per layer, single scattering, and 
    asymmetry averaged over the user input wave_range. 

    Parameters
    ----------
    out : list or dict 
        Either a list of output dictionaries or a single dictionary output
        from .compute(as_dict=True)
    wave_range : list 
        min and max wavelength in microns 
    return_output : bool 
        Default is just to return a figure but you can also 
        return all the 1d profiles 
    legend : bool 
        Default is none. Legend for each component of out 
    **kwargs : keyword arguments
        Key word arguments will be supplied to each bokeh figure function
    """
    kwargs['height'] = kwargs.get('plot_height',kwargs.get('height',300))
    kwargs['width'] = kwargs.get('plot_width', kwargs.get('width',300))
    if 'plot_width' in kwargs.keys() : kwargs.pop('plot_width')
    if 'plot_height' in kwargs.keys() : kwargs.pop('plot_height')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')

    if not isinstance(out, list):
        out = [out]

    pressure = out[0]['pressure']

    kwargs['y_range'] = kwargs.get('y_range',[max(pressure),min(pressure)])     

    ssa = figure(x_axis_label='Single Scattering Albedo',**kwargs)

    g0 = figure(x_axis_label='Asymmetry',**kwargs)

    opd = figure(x_axis_label='Optical Depth',y_axis_label='Pressure (bars)',
        x_axis_type='log',**kwargs)

    
    for i,results in enumerate(out): 
        inds = np.where((results['wave']>wave_range[0]) & 
            (results['wave']<wave_range[1]))
        opd_dat = np.mean(results['opd_per_layer'][:,inds],axis=2)[:,0]
        opd.line(opd_dat, 
                 results['pressure'], color=colors[np.mod(i, len(colors))],line_width=3)
        g0_dat = np.mean(results['asymmetry'][:,inds],axis=2)[:,0]
        g0.line(g0_dat, 
                 results['pressure'], color=colors[np.mod(i, len(colors))],line_width=3)
        
        if isinstance(legend, type(None)):
            ssa_dat = np.mean(results['single_scattering'][:,inds],axis=2)[:,0]
            ssa.line(ssa_dat, 
                 results['pressure'], color=colors[np.mod(i, len(colors))],line_width=3)
        else:
            ssa_dat = np.mean(results['single_scattering'][:,inds],axis=2)[:,0]
            ssa.line(ssa_dat, 
                 results['pressure'], color=colors[np.mod(i, len(colors))],line_width=3,
                 legend_label=legend[i])
            ssa.legend.location='top_left'

    if return_output:   
        return gridplot([[opd,ssa,g0]]), [opd_dat,ssa_dat,g0_dat]
    else:   
        return gridplot([[opd,ssa,g0]])

def find_nearest_1d(array,value):
    #small program to find the nearest neighbor in a matrix
    ar , iar ,ic = np.unique(array,return_index=True,return_counts=True)
    idx = (np.abs(ar-value)).argmin(axis=0)
    if ic[idx]>1: 
        idx = iar[idx] + (ic[idx]-1)
    else: 
        idx = iar[idx]
    return idx

def pressure_fig(**kwargs):
    kwargs['y_range'] = kwargs.get('y_range',[1e2,1e-6])
    kwargs['height'] = kwargs.get('plot_height',kwargs.get('height',400))
    kwargs['width'] = kwargs.get('plot_width', kwargs.get('width',600))
    if 'plot_width' in kwargs.keys() : kwargs.pop('plot_width')
    if 'plot_height' in kwargs.keys() : kwargs.pop('plot_height')
    kwargs['x_axis_label'] = kwargs.get('x_axis_label','Temperature (K)')
    kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure (bars)')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')        
    fig = figure(**kwargs)
    return fig

def plot_format(df):
    """Function to reformat plots"""
    df.xaxis.axis_label_text_font='times'
    df.yaxis.axis_label_text_font='times'
    df.xaxis.major_label_text_font_size='14pt'
    df.yaxis.major_label_text_font_size='14pt'
    df.xaxis.axis_label_text_font_size='14pt'
    df.yaxis.axis_label_text_font_size='14pt'
    df.xaxis.major_label_text_font='times'
    df.yaxis.major_label_text_font='times'
    df.xaxis.axis_label_text_font_style = 'bold'
    df.yaxis.axis_label_text_font_style = 'bold'

def plot_fsed(pressure, z, scale_height, alpha, beta, epsilon=1e-2, pres_alpha=None, **kwargs):

    kwargs['height'] = kwargs.get('plot_height',kwargs.get('height',300))
    kwargs['width'] = kwargs.get('plot_width', kwargs.get('width',700))
    if 'plot_width' in kwargs.keys() : kwargs.pop('plot_width')
    if 'plot_height' in kwargs.keys() : kwargs.pop('plot_height')
    kwargs['x_axis_label'] = kwargs.get('x_axis_label','fsed')
    kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure (bars)')
    kwargs['x_axis_type'] = kwargs.get('x_axis_type','log')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')

    if type(alpha) is int:
        alpha = [alpha]
    if type(beta) is int:
        beta = [beta]

    if pres_alpha is None:
        zstar = max(z)
    else:
        indx = find_nearest_1d(pressure, pres_alpha)
        zstar = z[indx]
    indx = find_nearest_1d(pressure, 1)
    H = scale_height[indx]

    cols = Colorblind8[:len(alpha)*len(beta)]
    lines = ['solid','dashed','dotted','dotdash','dashdot']
    kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure), np.min(pressure)])
    fig = figure(**kwargs)
    
    for i in range(len(alpha)):
        for j in range(len(beta)):
            lab = "alpha=%g" %alpha[i] + ", beta=%g" %beta[j]
            col = Colorblind8[np.mod(i+len(alpha)*j, 8)]
            line = lines[np.mod(j, 5)]
            fsed = alpha[i] * np.exp((z-zstar)/(6*beta[j]*H)) + epsilon
            fig.line(fsed, pressure, legend_label=lab, color=col, line_width=5, line_dash='solid')#line)

    fig.legend.location = "bottom_right"

    return fig

def fsed_from_output(out,labels,y_axis='pressure',color_indx=0,cld_bounds=False,gas_indx=None,**kwargs):

    if type(out)==dict:
        out=[out]

    kwargs['height'] = kwargs.get('plot_height',kwargs.get('height',300))
    kwargs['width'] = kwargs.get('plot_width', kwargs.get('width',700))
    if 'plot_width' in kwargs.keys() : kwargs.pop('plot_width')
    if 'plot_height' in kwargs.keys() : kwargs.pop('plot_height')
    kwargs['x_axis_label'] = kwargs.get('x_axis_label','fsed')
    kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure (bars)')
    kwargs['x_axis_type'] = kwargs.get('x_axis_type','log')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')
    #kwargs['x_range'] = kwargs.get('x_range', [1e-2, 2e1])
    if gas_indx is not None:
        title = 'Condensible = ' + str(out[0]['condensibles'][gas_indx])
        kwargs['title'] = kwargs.get('title', title)


    cols = Colorblind8[color_indx:color_indx+len(out)]
    pressure = out[0]['pressure']
    kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure), np.min(pressure)])
    fig = figure(**kwargs)

    min_id=[]; max_id=[]
    for i in range(len(out)):
        if gas_indx is None: x = out[i]['fsed']
        else: x = out[i]['fsed'][:,gas_indx]

        if y_axis == 'pressure':
            y = out[i]['pressure']
        elif y_axis == 'z':
            y = out[i]['altitude']
        col = Colorblind8[np.mod(i+color_indx, 8)]
        fig.line(x, y, legend_label=labels[i], color=col, line_width=5)
        if cld_bounds and gas_indx is not None:
            low_clds = []; high_clds = []
            ndz = out[i]['column_density'][:,gas_indx]
            nonzero = np.where(ndz>1e-3)[0]
            min_id.append(nonzero[0])
            max_id.append(nonzero[-1])
        
    if cld_bounds and gas_indx is not None:
        xmin = kwargs['x_range'][0]
        xmax = kwargs['x_range'][1]
        x1 = np.linspace(xmin,xmax,10)
        y1 = x1*0+y[min(min_id)]
        y2 = x1*0+y[max(max_id)]
        fig.line(x1, y1, color='black',line_width=5, line_dash='dashed')
        fig.line(x1, y2, color='black',line_width=5, line_dash='dashed')


    if labels is not None:
        fig.legend.location = "bottom_left"
    plot_format(fig)
    return fig

def aggregates_optical_properties(aggregate, mieff_dir, d_f_list, min_wavelength, max_wavelength):

    '''
    Code to plot the optical properties calculated in the .mieff files. for all radii and fractal dimensions analysed.

    Plots a 3D graph of Q_ext against wavelength (um) for all fractal dimensions (third axis).

    Set filtered = True if you want a smoother curve (using a savgol filter) for easier comparison of different fractals.

    Parameters
    ----------
    aggregate: string
        chemical species used in model 
    mieff_dir: string
        filepath to directory where .mieff files are stored
    d_f_list: array
        list of fractal dimensions analysed
    min_wavelength: float
        minimum wavelength to plot
    max_wavelength: float
        maximum wavelength to plot

    Returns
    -------
    3D graph of Q_ext against wavelength (um) for all fractal dimensions (third axis).

    '''

    # choose a colormap (based on the number of shapes to analyse)
    colors = plt.cm.winter_r(np.linspace(0,1,len(d_f_list)))


    # -------- SPHERES -------- 

    # read in the dataset for spherical particles
    mieff_col_names= ['A', 'B', 'C', 'D'] # assign arbitrary column names for reading in the data (all the columns have different numbers of rows)
    mieff_data = pd.read_csv(f"{mieff_dir}/{aggregate}.mieff", names=mieff_col_names, header=None, sep=r'\s+')

    num_wavelengths = int(mieff_data['A'][0]) # extract number of wavelengths from header
    num_radii = int(mieff_data['B'][0]) # extract number of radii from header

    # initialise lists
    wavelength_list_spheres=[]
    Q_ext_list_spheres=[]

    # Go through data for one radius size at a time, extracting the Q_ext and wavelength values and storing them in their respective lists
    for r in range(num_radii):

        wavelengths=[] # reset lists of wavelength and Q_ext each time we move onto a new radius
        Q_ext=[]
        
        for i in range(num_wavelengths):
            if (mieff_data['A'][r*num_wavelengths+r+i+2]<=max_wavelength) and (mieff_data['A'][r*num_wavelengths+r+i+2]>=min_wavelength): # only add this wavelength and Q_ext pair to our lists if it's smaller than the user-defined maximum
                wavelengths.append(mieff_data['A'][r*num_wavelengths+r+i+2]) # find the individual wavelengths at this radius
                Q_ext.append(mieff_data['C'][r*num_wavelengths+r+i+2]) # find the individual Q_ext values at each wavelength and at this radius (here the index 'r*num_wavelengths+r+i+2' basically filters the .mieff file to return it to a simple 2D array of "Wavelength, Qsca, Qext, g_asymmetry" for one particular particle size)

        wavelength_list_spheres.append(wavelengths)
        Q_ext_list_spheres.append(Q_ext)


    # record the radii used in the spherical .mieff file (not needed for the graph, but useful to print alongside it)
    radii=[]
    for r in range(num_radii):
        radii.append(10000.0*mieff_data['A'][r*num_wavelengths+r+1]) # convert all radii from cm (default for the .mieff file) to um as we record them


    # -------- AGGREGATES -------- 

    wavelength_list_of_lists=[]
    Q_ext_list_of_lists=[]

    for j in range(len(d_f_list)):

        # read in each dataset for aggregate particles one fractal dimension at a time
        mieff_data = pd.read_csv(f"{mieff_dir}/{aggregate}_aggregates_Df_{d_f_list[j]:.6f}.mieff", names=mieff_col_names, header=None, sep=r'\s+')

        # reset lists each fractal dimension
        wavelength_list=[]
        Q_ext_list=[]

        # Go through data for one radius size at a time, extracting the Q_ext and wavelength values and storing them in their respective lists
        for r in range(num_radii):

            wavelengths=[] # reset lists of wavelength and Q_ext each time we move onto a new radius
            Q_ext=[]
            
            for i in range(num_wavelengths):
                if (mieff_data['A'][r*num_wavelengths+r+i+2]<=max_wavelength) and (mieff_data['A'][r*num_wavelengths+r+i+2]>=min_wavelength): # only add this wavelength and Q_ext pair to our lists if it's smaller than the user-defined maximum
                    wavelengths.append(mieff_data['A'][r*num_wavelengths+r+i+2]) # find the individual wavelengths at this radius
                    Q_ext.append(mieff_data['C'][r*num_wavelengths+r+i+2]) # find the individual Q_ext values at each wavelength and at this radius (here the index 'r*num_wavelengths+r+i+2' basically filters the .mieff file to return it to a simple 2D array of "Wavelength, Qsca, Qext, g_asymmetry" for one particular particle size)

            wavelength_list.append(wavelengths)
            Q_ext_list.append(Q_ext)

        # save the lists of wavelengths and Q_ext for each fractal dimension
        wavelength_list_of_lists.append(wavelength_list)
        Q_ext_list_of_lists.append(Q_ext_list)


    # Create 3D scatterplot figure
    fig = plt.figure() 
    fig.set_facecolor('w') # make area around plot white in case user is in dark mode
    ax = plt.axes(projection ="3d")

    radii_colours = plt.cm.winter(np.linspace(0,1,len(wavelength_list_spheres))) # plot each radius in a different colour, from blue (small) to green (large)

    # SPHERES: plot each Q_ext vs wavelength graph, one radius (each list element) at a time 
    for i in range(len(wavelength_list_spheres)):
        ax.plot(wavelength_list_spheres[i], np.full(len(wavelength_list_spheres[i]),3), Q_ext_list_spheres[i], color = radii_colours[i], linewidth=0.5) # plot Q_ext for spheres at d_f = 3 (the np.full is just to make the array the same length, but the value for d_f is constant every time we run this line)

    # AGGREGATES: plot each Q_ext vs wavelength graph, one radius (each list element) and one fractal dimension at a time 
    for j in range(len(d_f_list)):
        for i in range(len(wavelength_list_spheres)):
            ax.plot(wavelength_list_of_lists[j][i], np.full(len(wavelength_list_of_lists[j][i]),d_f_list[j]), Q_ext_list_of_lists[j][i], color = radii_colours[i], linewidth=0.5) # plot Q_ext for each aggregate at a constant d_f value (the np.full is just to make the array the same length, but the value for d_f is constant every time we run this line)


    # set axes titles for 3D scatterplot
    ax.set_xlabel('$\lambda$ ($\mu$m)', fontsize=20)
    ax.set_ylabel('$d_f$', fontsize=20)
    ax.set_zlabel('$Q_{ext}$', fontsize=20)

    plt.close() # prevents duplicate plots from appearing

    return fig


def aggregates_wavelength_radius_grid(clouds_from_virga_spheres, clouds_from_virga_fractals, aggregate, d_f_list, mieff_dir, colors=None):

    '''

    Code to plot the wavelength-radius grid that was used the create the .mieff files, and then overplot the actual radii
    for each pressure layer for each fractal dimension, so that we can check that the radii we are using actually have
    unique optical properties that were calculated at a reasonable resolution.

    Parameters
    ----------
    clouds_from_virga_spheres: dict
        output from virga (spherical model)
    clouds_from_virga_fractals: dict
        array of dicts with output from virga (all of the virga models for each fractal dimension)
    aggregate: string
        chemical species used in model 
    d_f_list: array
        list of fractal dimensions analysed
    mieff_dir: string
        filepath to directory where .mieff files are stored
    colors: array (optional)
        list of colors to use (in order: spheres, 1.2, 1.6, 2.0, 2.4, 2.8)

    Returns
    -------
    Wavelength-radius grid from .mieff files, overplotted with radii for each pressure layer for each fractal dimension.

    '''

    # if colors = None, choose a colormap (based on the number of shapes to analyse).
    if(colors is None):
        color_array = plt.cm.viridis(np.linspace(0,1,len(d_f_list)+1))
    else:
        color_array = colors # or use colors provided by user


    # -------- SPHERES -------- 

    # read in the dataset for spherical particles
    mieff_col_names= ['A', 'B', 'C', 'D'] # assign arbitrary column names for reading in the data (all the columns have different numbers of rows)
    mieff_data = pd.read_csv(f"{mieff_dir}/{aggregate}.mieff", names=mieff_col_names, header=None, sep=r'\s+')

    num_wavelengths = int(mieff_data['A'][0]) # extract number of wavelengths from header
    num_radii = int(mieff_data['B'][0]) # extract number of radii from header

    # initialise lists
    list_of_sphere_wavelengths=[]
    list_of_sphere_radii=[]

    # record the actual wavelengths used in the spherical .mieff file
    for i in range(num_wavelengths):
        list_of_sphere_wavelengths.append(mieff_data['A'][i+2])

    # record the actual radii used in the spherical .mieff file
    for i in range(num_radii):
        list_of_sphere_radii.append(10000.0*mieff_data['A'][i*num_wavelengths+i+1]) # convert all radii from cm (default for the .mieff file) to um as we record them

    print(f'For the {aggregate}.mieff file (spheres):')
    print(f'\n\t Wavelengths are between {list_of_sphere_wavelengths[num_wavelengths-1]:.3f} --> {list_of_sphere_wavelengths[0]:.3f} um in {len(list_of_sphere_wavelengths)} intervals.')
    print(f'\t Radii are between {list_of_sphere_radii[0]} --> {list_of_sphere_radii[num_radii-1]} um in {len(list_of_sphere_radii)} intervals.')

    #print(list_of_sphere_radii)

    # load the data for the particle radius and column_density values for spheres
    column_density_data_spheres = clouds_from_virga_spheres['column_density'][(clouds_from_virga_spheres['column_density'] != 0.0).all(1)] # find the rows of data where we have clouds (where column_density was not equal to 0.0)
    radius_data_spheres = clouds_from_virga_spheres['mean_particle_r'].flatten(order='C')[(clouds_from_virga_spheres['column_density'] != 0.0).all(1)] # find the rows of data where we have clouds (where column_density was not equal to 0.0)



    # -------- AGGREGATES -------- 

    radius_data_aggregates_list=[] # initialise list to hold all of the radius values (in the cloud only) each fractal dimension
    column_density_data_aggregates_list=[] # initialise list to hold all of the column_density values (in the cloud only) each fractal dimension

    for j in range(len(d_f_list)):

        # ---- SAFETY CHECK -----

        # first, check that we used the same wavelength-radius grid as the spherical model for each fractal dimension (we don't need to use the data from these .mieff files again otherwise -- this is just a safety check)
        mieff_data = pd.read_csv(f"{mieff_dir}/{aggregate}_aggregates_Df_{d_f_list[j]:.6f}.mieff", names=mieff_col_names, header=None, sep=r'\s+') # load data

        # extract number of wavelenths and radii used in the grid
        num_wavelengths_aggregate = int(mieff_data['A'][0]) # extract number of wavelengths from header
        num_radii_aggregate = int(mieff_data['B'][0]) # extract number of radii from header

        # initialise lists
        list_of_aggregate_wavelengths=[]
        list_of_aggregate_radii=[]

        # record the actual wavelengths used in the aggregate .mieff file
        for i in range(num_wavelengths_aggregate):
            list_of_aggregate_wavelengths.append(mieff_data['A'][i+2])

        # record the actual radii used in the aggregate .mieff file
        for i in range(num_radii_aggregate):
            list_of_aggregate_radii.append(10000.0*mieff_data['A'][i*num_wavelengths_aggregate+i+1]) # convert all radii from cm (default for the .mieff file) to um as we record them

        # check wavelength grid is the same as for spheres

        for i in range(len(list_of_aggregate_wavelengths)):
            if(abs(list_of_aggregate_wavelengths[i] - list_of_sphere_wavelengths[i])>0.000001): # if the list of wavelengths is not equal to the version used in the spherical .mieff file, to 6 decimal places
                print('\n\n WARNING: Wavelength grid is not the same (to 6 dp)!') # print a warning
                print(f"\n Spherical version: ({num_wavelengths} wavelengths) does not match grid for aggregate {d_f_list[j]:.6f} ({int(mieff_data['A'][0])} wavelengths):\n") # explain where the error is
                print(f"\n The exact error occurs for SPHERE wavelength: {list_of_sphere_wavelengths[i]} um, where the AGGREGATE version is: {list_of_aggregate_wavelengths[i]} um.")
                print('\n Exiting code.\n')
                sys.exit() # exit the code

        # check radius grid is the same as for spheres
        for i in range(len(list_of_aggregate_radii)):
            if(abs(list_of_aggregate_radii[i] - list_of_sphere_radii[i])>0.000001): # if the list of radii is not equal to the version used in the spherical .mieff file, to 6 decimal places
                print('\n\n WARNING: Radius grid is not the same (to 6 dp)!') # print a warning
                print(f"\n Spherical version: ({num_radii} radii) does not match grid for aggregate {d_f_list[j]:.6f} ({int(mieff_data['B'][0])} radii):\n") # explain where the error is
                print(f"\n The exact error occurs for SPHERE radius: {list_of_sphere_wavelengths[i]} um, where the AGGREGATE version is: {list_of_aggregate_wavelengths[i]} um.")
                print('\n Exiting code.\n')
                sys.exit() # exit the code'''

        # ---- SAFETY CHECK COMPLETE -----


        # load the cloud data for the particle radius and column_density values for each aggregate and find the pressures that had clouds
        reduced_column_density_data = clouds_from_virga_fractals[j]['column_density'][(clouds_from_virga_fractals[j]['column_density'] != 0.0).all(1)] # find the rows of data where we have clouds (where column_density was not equal to 0.0)
        reduced_radius_data = clouds_from_virga_fractals[j]['mean_particle_r'].flatten(order='C')[(clouds_from_virga_fractals[j]['column_density'] != 0.0).all(1)] # find the rows of data where we have clouds (where column_density was not equal to 0.0)
        reduced_pressure_data = clouds_from_virga_fractals[j]['pressure'][(clouds_from_virga_fractals[j]['column_density'] != 0.0).all(1)] # find the rows of data where we have clouds (where column_density was not equal to 0.0) - we only need one of these, as all aggregates are on the same pressure grid (so no need to add to a list below)

        column_density_data_aggregates_list.append(reduced_column_density_data) # add this data to a list (one for each fractal dimension)
        radius_data_aggregates_list.append(reduced_radius_data) # add this data to a list (one for each fractal dimension)
        

    # check whether any of the particle radii are outside of the wavelength-radius grid
    warning_flag=0
    for j in range(len(radius_data_aggregates_list)):
        for i in range(len(radius_data_aggregates_list[j])):
            if (radius_data_aggregates_list[j][i] < list_of_sphere_radii[0]) or (radius_data_aggregates_list[j][i] > list_of_sphere_radii[num_radii-1]): # if the radius for one of the fractal dimensions is less than the minimum radius or larger than the maximum for the grid (which we have already checked is the same as the one in the spherical version)...
                print(f"WARNING: {d_f_list[j]} has a radius off the grid ({radius_data_aggregates_list[j][i]:.6f} um, at pressure {reduced_pressure_data[i]:.6f} bar).\n") #...print a warning and highlight which fractal dimension and radius caused it
                warning_flag=1
    
    if(warning_flag==0):
        print("\nGood news - all aggregates are within the grid!")

    # Create scatterplot figure
    fig,ax = plt.subplots()
    fig.set_facecolor('w') # make area around plot white in case user is in dark mode

    ax2 = ax.twinx() # make second y-axis (on right - column_density)

    # make wavelength-radius grid    
    ax.vlines(list_of_sphere_radii, list_of_sphere_wavelengths[num_wavelengths-1], list_of_sphere_wavelengths[0],  colors='lightgrey', linewidth=0.5) # add vertical lines for each radius in grid

    edge_of_grid_color = 'red' # choose colour for box marking outside of grid
    ax.vlines(list_of_sphere_radii[0], list_of_sphere_wavelengths[num_wavelengths-1], list_of_sphere_wavelengths[0],  colors=edge_of_grid_color)
    ax.vlines(list_of_sphere_radii[num_radii-1], list_of_sphere_wavelengths[num_wavelengths-1], list_of_sphere_wavelengths[0],  colors=edge_of_grid_color)
    ax.hlines(list_of_sphere_wavelengths[0], list_of_sphere_radii[0], list_of_sphere_radii[num_radii-1], colors=edge_of_grid_color)
    ax.hlines(list_of_sphere_wavelengths[num_wavelengths-1], list_of_sphere_radii[0], list_of_sphere_radii[num_radii-1], colors=edge_of_grid_color)


    marker_size = 15 # set marker size

    # SPHERES: add scatter plot of radii and 'column_density' for each pressure layer of cloud
    ax2.scatter(radius_data_spheres, column_density_data_spheres, color = color_array[0], label= f'Spheres', s=marker_size)

    # AGGREGATES: add scatter plot of radii and 'column_density' for each pressure layer of cloud
    for j in range(len(d_f_list)):
        ax2.scatter(radius_data_aggregates_list[j], column_density_data_aggregates_list[j], color = color_array[j+1], label= f'{d_f_list[j]:.1f}', s=marker_size)

    # make log-log plot
    ax.set_xscale('log')
    ax.set_yscale('log')
    ax2.set_yscale('log')

    # set axes titles
    ax.set_xlabel(r'Compact particle $r$ ($\mu$m)', fontsize=23)
    ax.set_ylabel(r'$\lambda$ ($\mu$m)', fontsize=23)
    ax2.set_ylabel(r'Column Density in layer ($\mathrm{cm}^{-2}$)', fontsize=23)

    # set tick font size
    ax.tick_params(axis='both', which='major', labelsize=20)
    ax.tick_params(axis='both', which='minor', labelsize=20)
    ax2.tick_params(axis='both', which='major', labelsize=20)
    ax2.tick_params(axis='both', which='minor', labelsize=20)
        
    #plt.legend(loc='lower left', bbox_to_anchor=(0.1, 0.1), prop={'size': 18})

    plt.close() # prevents duplicate plots from appearing

    return fig
    

def aggregates_pressure_vs_number_density(clouds_from_virga_spheres, clouds_from_virga_fractals, d_f_list, colors=None, min_pressure=None, max_pressure=None):

    '''

    Code to plot the particle densities at each radius and at each pressure layer for aggregates of each fractal dimension.
    This basically allows you to see the relative numbers and radii of the particles that form in each layer, for each fractal dimension.

    Parameters
    ----------
    clouds_from_virga_spheres: dict
        output from virga (spherical model)
    clouds_from_virga_fractals: dict
        array of dicts with output from virga (all of the virga models for each fractal dimension)
    d_f_list: array
        list of fractal dimensions analysed
    colors: array (optional)
        list of colors to use (in order: spheres, 1.2, 1.6, 2.0, 2.4, 2.8)
    min_pressure: float (optional)
        minimum pressure (bars) to plot on y-axis
    max_pressure: float (optional)
        maximum pressure (bars) to plot on y-axis

    Returns
    -------
    Figure with the particle densities at each radius and at each pressure layer for aggregates of each fractal dimension

    '''

    
    # if colors = None, choose a colormap (based on the number of shapes to analyse).
    if(colors is None):
        color_array = plt.cm.viridis(np.linspace(0,1,len(d_f_list)+1))
    else:
        color_array = colors # or use colors provided by user

    # set maximum and minimum marker sizes (on the screen) - these are arbitrary, and we can scale them to whatever looks good visually
    max_marker_size = 1000
    min_marker_size = 1

    # initialise a list to hold all of the data
    radius_data_list=[]

    # read in all datasets into a single list
    for i in range(len(d_f_list)):
        radius_data_list.extend(clouds_from_virga_fractals[i]['mean_particle_r'].flatten(order='C')) # add each fractal
    radius_data_list.extend(clouds_from_virga_spheres['mean_particle_r'].flatten(order='C')) # add spheres

    # find th min/max values that we will need to display on-screen
    global_min_r = min(n for n in radius_data_list if n>0) # record the minimum particle size (that is still above zero)
    global_max_r = max(radius_data_list) # record the maximum particle size

    #print(f'\nFINAL RESULTS: Min = {global_min_r}, Max = {global_max_r}')

    # calculate min and max r2 values because these relate linearly to the marker sizes, which are areas
    min_r_squared = global_min_r**2
    max_r_squared = global_max_r**2

    # scale the min/max radii so that they correspond with the min/max marker sizes that we want to show on screen
    marker_size_per_square_radii = (max_marker_size - min_marker_size) / (max_r_squared - min_r_squared)

    # Create scatterplot figure
    fig,ax = plt.subplots()
    fig.set_facecolor('w') # make area around plot white in case user is in dark mode

    # plot values for each fractal dimension d_f
    for i in range(len(d_f_list)):
        # create a list of marker sizes for the radii in this fractal dimension
        square_radii = clouds_from_virga_fractals[i]['mean_particle_r'].flatten(order='C')**2 # first, simply square the radii
        marker_sizes = min_marker_size + marker_size_per_square_radii * (square_radii - min_r_squared) # use derived equation to find the marker area that would give accurately scaled radii on screen
        # make scatter plot 
        legend_handle=ax.scatter(clouds_from_virga_fractals[i]['column_density'], clouds_from_virga_fractals[i]['pressure'], color = color_array[i+1], s=marker_sizes, label= f'{d_f_list[i]:.1f}') # legend handle is just there to make the legend markers all the same size (see it's use below)

    # plot values for spheres
    # create a list of marker sizes for the radii for spheres
    square_radii = clouds_from_virga_spheres['mean_particle_r'].flatten(order='C')**2 # first, simply square the radii
    marker_sizes = min_marker_size + marker_size_per_square_radii * (square_radii - min_r_squared) # use derived equation to find the marker area that would give accurately scaled radii on screen

    ax.scatter(clouds_from_virga_spheres['column_density'], clouds_from_virga_spheres['pressure'], color = color_array[0], s=marker_sizes, label= f'Spheres')

    # make log-log plot
    ax.set_yscale('log')
    ax.set_xscale('log')

    if (min_pressure is not None) and (max_pressure is not None):
        # if provided, set limits on y-axis (to begin at the cloud deck and above)
        ax.set_ylim(min_pressure, max_pressure)

    # invert the y-axis so that pressure decreases as you go upwards (basically representing altitude)
    plt.gca().invert_yaxis()

    # set axes titles
    ax.set_xlabel(r'Column Density in layer ($\mathrm{cm}^{-2}$)', fontsize=20)
    ax.set_ylabel(r'$P$ (bar)', fontsize=20)

    # set tick font size
    ax.tick_params(axis='both', which='major', labelsize=15)
    ax.tick_params(axis='both', which='minor', labelsize=15)

        
    # produce nice equal marker sizes in the legend
    handles, labels = ax.get_legend_handles_labels()
    ax.legend(list(reversed(handles)), list(reversed(labels)), title='$d_f$', title_fontsize=12, fontsize=12, loc='lower right')
    marker_size = 36
    def update_prop(handle, orig):
        handle.update_from(orig)
        handle.set_sizes([marker_size])
    plt.legend(handler_map={type(legend_handle): HandlerPathCollection(update_func=update_prop)})
    
    plt.close() # prevents duplicate plots from appearing

    return fig

def aggregates_pressure_vs_radius(clouds_from_virga_spheres, clouds_from_virga_fractals, d_f_list, colors=None, width=3, min_pressure=None, max_pressure=None):
    
    '''
    
    Code to plot the particle radius at each pressure layer for aggregates of each fractal dimension.
    Designed originally for diagnostic tests as part for VIRGA development by MGL.

    Parameters
    ----------
    clouds_from_virga_spheres: dict
        output from virga (spherical model)
    clouds_from_virga_fractals: dict
        array of dicts with output from virga (all of the virga models for each fractal dimension)
    d_f_list: array
        list of fractal dimensions analysed
    colors: array (optional)
        list of colors to use (in order: spheres, 1.2, 1.6, 2.0, 2.4, 2.8)
    width: float (optional)
        width of lines in plot (default = 3)
    min_pressure: float (optional)
        minimum pressure (bars) to plot on y-axis
    max_pressure: float (optional)
        maximum pressure (bars) to plot on y-axis
    
    Returns
    -------
    Figure plotting the particle radius at each pressure layer for aggregates of each fractal dimension.

    '''

    # if colors = None, choose a colormap (based on the number of shapes to analyse).
    if(colors is None):
        color_array = plt.cm.viridis(np.linspace(0,1,len(d_f_list)+1))
    else:
        color_array = colors # or use colors provided by user
        
    # Create scatterplot figure
    fig,ax = plt.subplots()
    fig.set_facecolor('w') # make area around plot white in case user is in dark mode

    # plot values for each fractal dimension d_f
    for i in range(len(d_f_list)):
        ax.plot(clouds_from_virga_fractals[i]['mean_particle_r'].flatten(order='C'), clouds_from_virga_fractals[i]['pressure'], color = color_array[i+1], label= f'{d_f_list[i]:.1f}', linewidth=width)

    # plot values for spheres
    ax.plot(clouds_from_virga_spheres['mean_particle_r'].flatten(order='C'), clouds_from_virga_spheres['pressure'], color = color_array[0], label= f'Spheres', linewidth=width)

    # make log-log plot
    ax.set_yscale('log')
    ax.set_xscale('log')

    if (min_pressure is not None) and (max_pressure is not None):
        # if provided, set limits on y-axis (to begin at the cloud deck and above)
        ax.set_ylim(min_pressure, max_pressure)

    # invert the y-axis so that pressure decreases as you go upwards (basically representing altitude)
    plt.gca().invert_yaxis()

    # set axes titles
    ax.set_xlabel(r'Compact particle $r$ ($\mu$m)', fontsize=20)
    ax.set_ylabel(r'$P$ (bar)', fontsize=20)

    # set tick font size
    ax.tick_params(axis='both', which='major', labelsize=15)
    ax.tick_params(axis='both', which='minor', labelsize=15)

    plt.legend(prop={'size': 12})

    plt.close() # prevents duplicate plots from appearing

    return fig



import numpy as np

def TiO2(mw_atmos, mh=1, gas_mmr = None):
    """Defines properties for TiO2 as condensible
    
    Parameters 
    ----------
    mw_atmos : float 
        Mean molecular weight of the atmosphere amu
    gas_mmr : float , optional
        Gas mass mixing ratio.
        None points to the default value of : 1.69e-7
    mh : float , optional
        Metallicity, Default is 1=1xSolar
    
    Returns
    -------
    mean molecular weight of gas,
    gas mass mixing ratio 
    density of gas cgs
    """
    #if mh != 1: raise Exception("Alert: No M/H Dependence in TiO2 Routine. Consult your local theorist to determine next steps.")
    if isinstance(gas_mmr, type(None)):
        gas_mmr = 1.69e-7 * mh
    gas_mw = 80
    gas_mmr = gas_mmr * (gas_mw/mw_atmos) 
    rho_p =  4.25
    return gas_mw, gas_mmr, rho_p

def CH4(mw_atmos , mh = 1,gas_mmr = None):
    """Defines properties for CH4 as condensible
    
    Parameters 
    ----------
    mw_atmos : float 
        Mean molecular weight of the atmosphere amu
    gas_mmr : float , optional
        Gas mass mixing ratio.
        None points to the default value of : 4.9e-4
    mh : float , optional
        Metallicity, Default is 1=1xSolar
    
    Returns
    -------
    mean molecular weight of gas,
    gas mass mixing ratio 
    density of gas cgs
    """
    #if mh != 1: raise Exception("Alert: No M/H Dependence in CH4 Routine. Consult your local theorist to determine next steps.")
    #ch4 is linearly dependent on m/h at cold temperatures but scales as m^0 at hotter temps where co is dominant 
    #but we can assume a linear dependence here 
    if isinstance(gas_mmr, type(None)):
        gas_mmr = 4.9e-4 * mh

    gas_mw = 16.0
    # Lodders(2003)
    gas_mmr = gas_mmr * (gas_mw/mw_atmos) 
    #V.G. Manzhelii and A.M. Tolkachev, Sov. Phys. Solid State 5, 2506 (1964)
    rho_p =  0.49   #solid (T = 213 K)
    return gas_mw, gas_mmr, rho_p

def NH3(mw_atmos, mh = 1, gas_mmr = None):
    """Defines properties for NH3 as condensible
    
    Parameters 
    ----------
    mw_atmos : float 
        Mean molecular weight of the atmosphere amu
    gas_mmr : float , optional
        Gas mass mixing ratio
        None points to the default value of : 1.34e-4
    mh : float , optional
        Metallicity, Default is 1=1xSolar
    
    Returns
    -------
    mean molecular weight of gas,
    gas mass mixing ratio 
    density of gas cgs
    """
    #if mh != 1: raise Exception("Alert: No M/H Dependence in NH3 Routine. Consult your local theorist to determine next steps.")
    if isinstance(gas_mmr, type(None)):
        gas_mmr = 1.34e-4 * mh

    gas_mw = 17.
    #Lodders(2003)
    gas_mmr = gas_mmr * (gas_mw/mw_atmos) 
    #V.G. Manzhelii and A.M. Tolkachev, Sov. Phys. Solid State 5, 2506 (1964)
    rho_p =  0.84  #solid (T = 213 K)
    return gas_mw, gas_mmr, rho_p

def H2O(mw_atmos, mh = 1, gas_mmr = None):
    """Defines properties for H2O as condensible
    
    Parameters 
    ----------
    mw_atmos : float 
        Mean molecular weight of the atmosphere amu
    gas_mmr : float , optional
        Gas mass mixing ratio
        None points to the default value of : 7.54e-4
    mh : float , optional
        Metallicity, Default is 1=1xSolar
    
    Returns
    -------
    mean molecular weight of gas,
    gas mass mixing ratio 
    density of gas cgs
    """
    #if mh != 1: raise Exception("Alert: No M/H Dependence in H2O Routine. Consult your local theorist to determine next steps.")
    if isinstance(gas_mmr, type(None)):
        gas_mmr = 7.54e-4 * mh
    gas_mw = 18.
    #Lodders(2003)
    gas_mmr = gas_mmr * (gas_mw/mw_atmos)  
    #V.G. Manzhelii and A.M. Tolkachev, Sov. Phys. Solid State 5, 2506 (1964)
    rho_p =  0.93   #solid (T = 213 K)
    return gas_mw, gas_mmr, rho_p

def Fe(mw_atmos,  mh = 1,gas_mmr = None):
    """Defines properties for Fe as condensible
    
    Parameters 
    ----------
    mw_atmos : float 
        Mean molecular weight of the atmosphere amu
    gas_mmr : float , optional
        Gas mass mixing ratio
        None points to the default value of : 5.095e-5
    mh : float , optional
        Metallicity, Default is 1=1xSolar
    
    Returns
    -------
    mean molecular weight of gas,
    gas mass mixing ratio 
    density of gas cgs

    Notes
    -----
    .. [1] Morley C.~V., Mukherjee S., Marley M.~S., Fortney J.~J., Visscher C., Lupu R., Gharib-Nezhad E., et al., 2024, ApJ, 975, 59. doi:10.3847/1538-4357/ad71d5

    """
    #if mh != 1: raise Exception("Alert: No M/H Dependence in Fe Routine. Consult your local theorist to determine next steps.")
    if isinstance(gas_mmr, type(None)):
        gas_mmr =  5.095e-5 * mh#5.78e-5 * mh #
    gas_mw = 55.845
    gas_mmr = gas_mmr * (gas_mw/mw_atmos)
    #Lodders and Fegley (1998)
    rho_p =  7.875  
    return gas_mw, gas_mmr, rho_p

def KCl(mw_atmos, mh = 1, gas_mmr = None):
    """Defines properties for KCl as condensible
    
    Parameters 
    ----------
    mw_atmos : float 
        Mean molecular weight of the atmosphere amu
    gas_mmr : float , optional
        Gas mass mixing ratio
    mh : float , optional
        Metallicity, Default is 1=1xSolar
    
    Returns
    -------
    Mean molecular weight of gas,
    Gas mass mixing ratio 
    Density of gas cgs
    """ 
    #if mh != 1: raise Exception("Alert: No M/H Dependence in Fe Routine. Consult your local theorist to determine next steps.")

    gas_mw = 74.5
    #mh dependence from Morley12
    #mh_values = np.array([0,0.3,0.5,0.7,1.0,1.7,2.0,2.5])
    #gas_mmrs = np.array([2.2627E-07, 2*2.2627E-07, 3.1*2.2627E-07,5*2.2627E-07, 2.1829E-06,  8.1164E-06,2*8.1164E-06, 3*2*8.1164E-06])

    if isinstance(gas_mmr, type(None)):
        gas_mmr = 2.55e-7 * mh #np.interp(np.log10(mh), mh_values, gas_mmrs) * (gas_mw/mw_atmos)
    
    gas_mmr = gas_mmr * (gas_mw/mw_atmos)
    #source unknown
    rho_p =  1.99
    return gas_mw, gas_mmr, rho_p

def MgSiO3(mw_atmos, mh = 1, gas_mmr = None):
    """Defines properties for MgSiO3 as condensible
    
    Parameters 
    ----------
    mw_atmos : float 
        Mean molecular weight of the atmosphere amu
    gas_mmr : float , optional
        Gas mass mixing ratio
        None points to the default value of : 2.75e-3
    mh : float , optional
        Metallicity, Default is 1=1xSolar
    
    Returns
    -------
    Mean molecular weight of gas,
    Gas mass mixing ratio 
    Density of gas cgs

    Notes
    -----
    .. [1] Morley C.~V., Mukherjee S., Marley M.~S., Fortney J.~J., Visscher C., Lupu R., Gharib-Nezhad E., et al., 2024, ApJ, 975, 59. doi:10.3847/1538-4357/ad71d5

    """ 
    gas_mw = 100.4
    #if mh != 1: raise Exception("Alert: No M/H Dependence in MgSiO3 Routine. Consult your local theorist to determine next steps.")
    if isinstance(gas_mmr, type(None)):
        gas_mmr =  2.92e-5 *  mh # # gao 202 nat astro 60.3e-6 
    
    gas_mmr =  gas_mmr * (gas_mw/mw_atmos)
    #Lodders and Fegley (1998)
    rho_p =  3.192
    return gas_mw, gas_mmr, rho_p

def Mg2SiO4(mw_atmos, mh = 1,gas_mmr = None,):
    """Defines properties for Mg2SiO4 as condensible
    
    Parameters 
    ----------
    mw_atmos : float 
        Mean molecular weight of the atmosphere amu
    gas_mmr : float , optional
        Gas mass mixing ratio
        None points to the default value of : 3.58125e-05
    mh : float , optional
        Metallicity, Default is 1=1xSolar
    
    Returns
    -------
    Mean molecular weight of gas,
    Gas mass mixing ratio 
    Density of gas cgs

    Notes
    -----
    .. [1] Morley C.~V., Mukherjee S., Marley M.~S., Fortney J.~J., Visscher C., Lupu R., Gharib-Nezhad E., et al., 2024, ApJ, 975, 59. doi:10.3847/1538-4357/ad71d5

    """ 
    #if mh != 1: raise Exception("Alert: No M/H Dependence in Mg2SiO4 Routine. Consult your local theorist to determine next steps.")
    if isinstance(gas_mmr, type(None)):
        gas_mmr =  3.063e-5 * mh #59.36e-6 * mh
    gas_mw = 140.7
    #NEW FORSTERITE (from Lodders et al. table, 1000mbar, 1900K)
    gas_mmr = gas_mmr * (gas_mw/mw_atmos)  
    #Lodders and Fegley (1998)
    rho_p =  3.214  
    return gas_mw, gas_mmr, rho_p

def MnS(mw_atmos, mh=1,gas_mmr = None): 
    """Defines properties for MnS as condensible
    
    Parameters 
    ----------
    mw_atmos : float 
        Mean molecular weight of the atmosphere amu
    gas_mmr : float , optional
        Gas mass mixing ratio
        None points to the default value of : 6.32e-7
    mh : float , optional
        Metallicity, Default is 1=1xSolar
    
    Returns
    -------
    Mean molecular weight of gas,
    Gas mass mixing ratio 
    Density of gas cgs

    Notes
    -----
    .. [1] Lodders, K. Solar System Abundances of the Elements. Astrophys. Space Sci.
Proc. 16, 379 (2010)
    """ 
    if isinstance(gas_mmr, type(None)):
        gas_mmr =  5.41e-7 * mh

    gas_mw = 87.00

    gas_mmr = gas_mmr * (gas_mw/mw_atmos) 

    #Lodders and Fegley (2003) (cvm)
    rho_p =  4.0
    return gas_mw, gas_mmr, rho_p

def ZnS(mw_atmos, mh=1, gas_mmr=None):
    """Defines properties for ZnS as condensible
    
    Parameters 
    ----------
    mw_atmos : float 
        Mean molecular weight of the atmosphere amu
    gas_mmr : float , optional
        Gas mass mixing ratio
        None points to the default value of : 8.40e-8
    mh : float , optional
        Metallicity, Default is 1=1xSolar
    
    Returns
    -------
    Mean molecular weight of gas,
    Gas mass mixing ratio 
    Density of gas cgs

    Notes
    -----
    .. [1] Lodders, K. Solar System Abundances of the Elements. Astrophys. Space Sci.
Proc. 16, 379 (2010)
    """ 
    if isinstance(gas_mmr, type(None)):
        gas_mmr =  0.076e-6 * mh #lodders 2010

    gas_mw = 97.46

    gas_mmr = gas_mmr * (gas_mw/mw_atmos) 

    #Lodders and Fegley (2003) (cvm)
    rho_p =  4.04   
    return gas_mw, gas_mmr, rho_p

def Cr(mw_atmos, mh=1,gas_mmr=None):
    """Defines properties for Cr as condensible
    
    Parameters 
    ----------
    mw_atmos : float 
        Mean molecular weight of the atmosphere amu
    gas_mmr : float , optional
        Gas mass mixing ratio
        None points to the default value of :
        1xSolar = 8.87e-7
        10xSolar = 8.6803E-06
        50xSolar = 4.1308E-05
    mh : float , optional
        Metallicity, Default is 1=1xSolar
    
    Returns
    -------
    Mean molecular weight of gas,
    Gas mass mixing ratio 
    Density of gas cgs

    Notes
    -----
    .. [1] Lodders, K. Solar System Abundances of the Elements. Astrophys. Space Sci.
    Proc. 16, 379 (2010)    
    """     
    gas_mw = 51.996
    if isinstance(gas_mmr, type(None)):
        gas_mmr = 8.87e-7 *  mh
        #if mh==1:
        #    gas_mmr = 8.87e-7 * (gas_mw/mw_atmos) 
        #elif mh==10:
        #    gas_mmr = 8.6803E-06 * (gas_mw/mw_atmos) 
        #elif mh==50:
        #    gas_mmr = 4.1308E-05 * (gas_mw/mw_atmos) 
        #else: 
        #    raise Exception("Chromium (Cr) gas properties are only available for 1, 10, and 50xSolar")
    
    gas_mmr = gas_mmr * (gas_mw/mw_atmos)   

    #Lodders and Fegley (2003) (cvm)
    rho_p =  7.15
    return gas_mw, gas_mmr, rho_p

    
def Al2O3(mw_atmos, mh = 1, gas_mmr = None):
    """Defines properties for Al2O3 as condensible
    
    Parameters 
    ----------
    mw_atmos : float 
        Mean molecular weight of the atmosphere amu
    gas_mmr : float , optional
        Gas mass mixing ratio
        None points to the default value of : 2.51e-6
    mh : float , optional
        Metallicity, Default is 1=1xSolar
    
    Returns
    -------
    Mean molecular weight of gas,
    Gas mass mixing ratio 
    Density of gas cgs

    Notes
    -----
    .. [1] Gao, Peter, et al. "Aerosol composition of hot giant exoplanets dominated by silicates and hydrocarbon hazes." Nature Astronomy (2020): 1-6.
    .. [2] Lodders, Katharina. "Solar system abundances of the elements." Principles and perspectives in cosmochemistry. Springer, Berlin, Heidelberg, 2010. 379-417.
    .. [3] Morley C.~V., Mukherjee S., Marley M.~S., Fortney J.~J., Visscher C., Lupu R., Gharib-Nezhad E., et al., 2024, ApJ, 975, 59. doi:10.3847/1538-4357/ad71d5
    """     
    #if mh != 1: raise Exception("Alert: No M/H Dependence in Al2O3 Routine. Consult your local theorist to determine next steps.")
    if isinstance(gas_mmr, type(None)):
        #gas_mmr = 4.937e-6 #gao 2020 value 
        gas_mmr = 2.489e-6 * mh #diamond back

    gas_mw = 101.961
    #NEW FORSTERITE (from Lodders et al. table, 1000mbar, 1900K)
    gas_mmr = gas_mmr * (gas_mw/mw_atmos)  
    #Lodders and Fegley (1998)
    rho_p =  3.987
    return gas_mw, gas_mmr, rho_p

def Na2S(mw_atmos, mh = 1, gas_mmr = None):
    """Defines properties for Na2S as condensible
    
    Parameters 
    ----------
    mw_atmos : float 
        Mean molecular weight of the atmosphere amu
    gas_mmr : float , optional
        Gas mass mixing ratio
        None points to the default value of : 3.97e-6 
    mh : float , optional
        Metallicity, Default is 1=1xSolar
    
    Returns
    -------
    Mean molecular weight of gas,
    Gas mass mixing ratio 
    Density of gas cgs

    Notes
    -----
    .. [1] Lodders, K. Solar System Abundances of the Elements. Astrophys. Space Sci.
Proc. 16, 379 (2010)
    """     
    #if mh != 1: raise Exception("Alert: No M/H Dependence in Na2S Routine. Consult your local theorist to determine next steps.")
    if isinstance(gas_mmr, type(None)):
        gas_mmr = 3.34e-6 * mh
    gas_mw = 78.05
    gas_mmr = gas_mmr * (gas_mw/mw_atmos)  
    #Lodders and Fegley (1998)
    rho_p =  1.856
    return gas_mw, gas_mmr, rho_p

def CaTiO3(mw_atmos, mh = 1, gas_mmr=None):
    """Defines properties for CaTiO3 as condensible
    
    Parameters 
    ----------
    mw_atmos : float 
        Mean molecular weight of the atmosphere amu
    gas_mmr : float , optional
        Gas mass mixing ratio
        None points to the default value of : 2.51e-6
    mh : float , optional
        Metallicity, Default is 1=1xSolar
    
    Returns
    -------
    Mean molecular weight of gas,
    Gas mass mixing ratio 
    Density of gas cgs
    """     
    #if mh != 1: raise Exception("Alert: No M/H Dependence in CaTiO3 Routine. Consult your local theorist to determine next steps.")
    if isinstance(gas_mmr, type(None)):
        gas_mmr = 2.51e-6 * mh 
    gas_mw = 135.745
    gas_mmr = gas_mmr * (gas_mw/mw_atmos)
    rho_p =  3.987
    return gas_mw, gas_mmr, rho_p

def CaAl12O19(mw_atmos, mh = 1,gas_mmr=None):
    """Defines properties for CaAl12O19 as condensible
    
    Parameters 
    ----------
    mw_atmos : float 
        Mean molecular weight of the atmosphere amu
    gas_mmr : float , optional
        Gas mass mixing ratio
        None points to the default value of : 8.80e-7
    mh : float , optional
        Metallicity, Default is 1=1xSolar
    
    Returns
    -------
    Mean molecular weight of gas,
    Gas mass mixing ratio 
    Density of gas cgs
    """
    #if mh != 1: raise Exception("Alert: No M/H Dependence in CaAl12O19 Routine. Consult your local theorist to determine next steps.")
    if isinstance(gas_mmr, type(None)):
        gas_mmr = 8.80e-7 * mh 
    gas_mw = 667.843
    gas_mmr = gas_mmr * (gas_mw/mw_atmos)
    rho_p =  7.15
    return gas_mw, gas_mmr, rho_p

def SiO2(mw_atmos, mh=1, gas_mmr = None):
    """Defines properties for TiO2 as condensible
    
    Parameters 
    ----------
    mw_atmos : float 
        Mean molecular weight of the atmosphere amu
    gas_mmr : float , optional
        Gas mass mixing ratio.
        None points to the default value of : 1.69e-7
    mh : float , optional
        Metallicity, Default is 1=1xSolar
    
    Returns
    -------
    mean molecular weight of gas,
    gas mass mixing ratio 
    density of gas cgs
    """
    #if mh != 1: raise Exception("Alert: No M/H Dependence in SiO2 Routine. Consult your local theorist to determine next steps.")
    if isinstance(gas_mmr, type(None)):
        gas_mmr = 60.3e-6 * mh
    gas_mw = 60
    gas_mmr = gas_mmr * (gas_mw/mw_atmos) 
    rho_p =  2.648
    return gas_mw, gas_mmr, rho_p


import numpy as np
import astropy.units as u

from virga import justdoit as jdi

def test_mie_database():
    qext, qsca, asym, radii, wave = jdi.calc_mie_db(['MnS'],
                                '.', '.', rmin = 1e-5, nradii = 10)

    assert np.isclose(np.sum(qext), 3943.0645661036983)
    assert np.isclose(np.sum(qsca), 3761.3770094896213)
    assert np.isclose(np.sum(asym), 1918.452490845249)

def test_virga_cloud():
    # initialise atmosphere
    a = jdi.Atmosphere(['MnS'], fsed=1, mh=1, mmw=2.2)
    a.gravity(gravity=7.460, gravity_unit=u.Unit('m/(s**2)'))
    a.ptk(df=jdi.hot_jupiter())

    # calculate cloud profile
    all_out = jdi.compute(a, as_dict=True, directory='.')

    assert np.isclose(np.sum(all_out['condensate_mmr']), 6.163947994805619e-05)
    
    # initialise atmosphere
    a = jdi.Atmosphere('MnS', fsed=1, mh=1, mmw=2.2)
    a.gravity(gravity=7.460, gravity_unit=u.Unit('m/(s**2)'))
    a.ptk(df=jdi.hot_jupiter())

    # calculate cloud profile
    all_out = jdi.compute(a, as_dict=True, directory='.')

    assert np.isclose(np.sum(all_out['condensate_mmr']), 6.163947994805619e-05)




import numpy as np
pi=np.pi
import miepython as mie
import os 
import subprocess
import pandas as pd
from .root_functions import calculate_k0, find_N_mon_and_r_mon

def fort_mie_calc(RO, RFR, RFI, THET, JX, R, RE2, TMAG2, WVNO ):
    """
    Given the refractive indices at a certain wavelength this module
    calculates the Mie scattering by a stratified sphere.The basic code used 
    was that described in the report: " Subroutines for computing the parameters of 
    the electromagnetic radiation scattered by a sphere " J.V. Dave,
    I B M Scientific Center, Palo Alto , California.
    Report NO. 320 - 3236 .. MAY 1968 .

    Parameters
    ----------
    RO : float
        Outer Shell Radius (cm)
    RFR : float
        Real refractive index of shell layer (in the form n= RFR-i*RFI)
    RFI : float
        Imaginary refractive index of shell layer (in the form n= RFR-i*RFI)
    THET : ndarray 
        Angle in degrees between the directions of the incident and the scattered radiation.
    JX : integer
        Total number of THET for which calculations are required
    R : float
        Radius of core (cm)`
    RE2 : float 
        Real refractive index of core (in the form n= RE2-i*TMAG2)
    TMAG2 : float
        Imaginary refractive index of core (in the form n= RE2-i*TMAG2)
        
    WVNO : float
        Wave-number corresponding to the wavelength. (cm^-1)
    
    Returns
    -------
    QEXT: float
        Efficiency factor for extinction,VAN DE HULST,P.14 ' 127
    QSCAT: float
        Efficiency factor for scattering,VAN DE HULST,P.14 ' 127
    CTBQRS: float
        Average(cos(theta))*QSCAT,VAN DE HULST,P.14 ' 127
    ISTATUS: integer
        Convergence indicator, 0 if converged, -1 if otherwise.
        
        
    """
    
    EPSILON_MIE = 1e-7  ## Tolerance for convergence

    nacap, IT = 1000000, 1
    
    ACAP=np.zeros(shape=(nacap),dtype=complex) 
    W=np.zeros(shape=(3,nacap),dtype=complex)
    WFN=np.zeros(shape=(2),dtype=complex)
    Z=np.zeros(shape=(4),dtype=complex)
    U=np.zeros(shape=(8),dtype=complex)
    T= np.zeros(shape=(5))    
    TA=np.zeros(shape=(4))     
    TB=np.zeros(shape=(2))        
    TC=np.zeros(shape=(2))
    TD=np.zeros(shape=(2))  
    TE=np.zeros(shape=(2))  
    PI=np.zeros(shape=(3,IT))   
    TAU=np.zeros(shape=(3,IT))
    CSTHT=np.zeros(shape=(IT)) 
    THETD=np.full((IT),THET) 
    SI2THT=np.zeros(shape=(IT))   
    ELTRMX=np.zeros(shape=(4,IT,2))
    IFLAG = 1
    if  R/RO < 1e-6:    
        IFLAG = 2
    if  JX > IT:
        raise Exception('THE VALUE OF THE ARGUMENT JX IS GREATER THAN IT. PLEASE READ COMMENTS.')
        
            
    RF =  complex(RFR,-RFI)
    RC =  complex( RE2,-TMAG2 )
    X  =  RO * WVNO
    K1 =  RC * WVNO
    K2 =  RF * WVNO
    K3 =  complex( WVNO, 0.0 )
    Z[0] =  K2 * RO
    Z[1] =  K3 * RO
    Z[2] =  K1 * R
    Z[3] =  K2 * R
    X1   =  Z[0].real 
    X4   =  Z[3].real
    Y1   =   Z[0].imag
    Y4   = Z[3].imag
    RRF  =  1.0 / RF
    RX   =  1.0 / X
    RRFX =  RRF * RX
    T[0] = np.sqrt( ( X**2 ) * ( RFR**2 + RFI**2 ))
    NMX1 = int( 1.10 * T[0])
    if  NMX1 > nacap-1 :
        istatus=-1
        return 0,0,0,istatus
        
    NMX2 = int(T[0])
    if  NMX1 <=  150 :
        NMX1 = 150
        NMX2 = 135
    
    ACAP[NMX1]  =  complex( 0.0,0.0 ) #+1
    if IFLAG != 2:
        for N in range(3):
            W[N,NMX1]  = complex( 0.0,0.0 ) #+1
    for N in range(NMX1):
        NN = NMX1 - N-1 ## removed a plus 1 to make up for python loop
        ACAP[NN] = (NN+2) * RRFX - 1.0 / ( (NN+2) * RRFX + ACAP[NN+1] )
        if  IFLAG != 2 :
            for M in range(3):
                W[ M,NN ] = ((NN+2) / Z[M+1])  -1.0 / (  (NN+2) / Z[M+1]  +  W[ M,NN+1 ]  )
    for J in range(JX):
        if  THETD[J] < 0.0 :
            THETD[J] =  abs( THETD[J] )
    
        if  THETD[J] < 90.0:
            T[0]     =  ( 3.14159265359 * THETD[J] ) / 180.0
            CSTHT[J] =  np.cos( T[0])
            SI2THT[J] =  1.0 - CSTHT[J]**2
            continue
        if  THETD[J] == 90.0:
            CSTHT[J]  =  0.0
            SI2THT[J] =  1.0
            continue
        if  THETD[J] > 90.0 :
            raise Exception(' THE VALUE OF THE SCATTERING ANGLE IS GREATER THAN 90.0 DEGREES. PLEASE READ COMMENTS')
        
    for J in range(JX):
        PI[0,J]  =  0.0
        PI[1,J]  =  1.0
        TAU[0,J] =  0.0
        TAU[1,J] =  CSTHT[J]
    T[0]   =  np.cos(X)
    T[1]   =  np.sin(X)
    WM1    =  complex( T[0],-T[1] )
    WFN[0] =  complex( T[1], T[0] )
    TA[0]  =  T[1]
    TA[1]  =  T[0]
    WFN[1] =  RX * WFN[0] - WM1
    TA[2]  =  (WFN[1].real)
    TA[3]  =  (WFN[1].imag)
    if IFLAG != 2:
        N=1
        SINX1   =  np.sin( X1 )
        SINX4   =  np.sin( X4 )
        COSX1   =  np.cos( X1 )
        COSX4   =  np.cos( X4 )
        EY1     =  np.exp( Y1 )
        E2Y1    =  EY1 * EY1
        EY4     =  np.exp( Y4 )
        EY1MY4  =  np.exp( Y1 - Y4 )
        EY1PY4  =  EY1 * EY4
        EY1MY4  =  np.exp( Y1 - Y4 )
        AA  =  SINX4 * ( EY1PY4 + EY1MY4 )
        BB  =  COSX4 * ( EY1PY4 - EY1MY4 )
        CC  =  SINX1 * ( E2Y1 + 1.0 )
        DD  =  COSX1 * ( E2Y1 - 1.0 )
        DENOM   =  1.0  +  E2Y1 * ( 4.0 * SINX1 * SINX1 - 2.0 + E2Y1 )
        REALP   =  ( AA * CC  +  BB * DD ) / DENOM
        AMAGP   =  ( BB * CC  -  AA * DD ) / DENOM
        DUMMY   =  complex( REALP, AMAGP )
        AA  =  SINX4 * SINX4 - 0.5
        BB  =  COSX4 * SINX4
        P24H24  =  0.5 + complex( AA,BB ) * EY4 * EY4
        AA  =  SINX1 * SINX4  -  COSX1 * COSX4
        BB  =  SINX1 * COSX4  +  COSX1 * SINX4
        CC  =  SINX1 * SINX4  +  COSX1 * COSX4
        DD  = -SINX1 * COSX4  +  COSX1 * SINX4
        P24H21  =  0.5 * complex( AA,BB ) * EY1 * EY4 + 0.5 * complex( CC,DD ) * EY1MY4
        DH4  =  Z[3] / ( 1.0 + complex( 0.0,1.0 ) * Z[3] )  -  1.0 / Z[3]
        DH1  =  Z[0] / ( 1.0 + complex( 0.0,1.0 ) * Z[0] )  -  1.0 / Z[0]
        DH2  =  Z[1] / ( 1.0 + complex( 0.0,1.0 ) * Z[1] )  -  1.0 / Z[1]
        PSTORE  =  ( DH4 + N / Z[3] )  *  ( W[2,N-1] + N / Z[3] )
        P24H24  =  P24H24 / PSTORE
        HSTORE  =  ( DH1 + N / Z[0] )  *  ( W[2,N-1] + N / Z[3] )
        P24H21  =  P24H21 / HSTORE
        PSTORE  =  ( ACAP[N-1] + N / Z[0] )  /  ( W[2,N-1] + N / Z[3] )
        DUMMY   =  DUMMY * PSTORE
        DUMSQ   =  DUMMY * DUMMY
        U[0] =  K3 * ACAP[N-1]  -  K2 * W[0,N-1]
        U[1] =  K3 * ACAP[N-1]  -  K2 * DH2
        U[2] =  K2 * ACAP[N-1]  -  K3 * W[0,N-1]
        U[3] =  K2 * ACAP[N-1]  -  K3 * DH2
        U[4] =  K1 *  W[2,N-1]  -  K2 * W[1,N-1]
        U[5] =  K2 *  W[2,N-1]  -  K1 * W[1,N-1]
        U[6] =  complex( 0.0,-1.0 )  *  ( DUMMY * P24H21 - P24H24 )
        U[7] =  TA[2] / WFN[1]
            
        FNA  =  U[7] * ( U[0]*U[4]*U[6]  +  K1*U[0]  -  DUMSQ*K3*U[4] ) /( U[1]*U[4]*U[6]  +  K1*U[1]  -  DUMSQ*K3*U[4] )
        FNB  =  U[7] * ( U[2]*U[5]*U[6]  +  K2*U[2]  -  DUMSQ*K2*U[5] ) /( U[3]*U[5]*U[6]  +  K2*U[3]  -  DUMSQ*K2*U[5] )
        TB[0]=FNA.real
        TB[1]=FNA.imag
        TC[0]=FNB.real
        TC[1]=FNB.imag
    elif IFLAG == 2:
        TC1  =  ACAP[0] * RRF  +  RX
        TC2  =  ACAP[0] * RF   +  RX
        FNA  =  ( TC1 * TA[2]  -  TA[0] ) / ( TC1 * WFN[1]  -  WFN[0] )
        FNB  =  ( TC2 * TA[2]  -  TA[0] ) / ( TC2 * WFN[1]  -  WFN[0] )
        TB[0]=FNA.real
        TB[1]=FNA.imag
        TC[0]=FNB.real
        TC[1]=FNB.imag
        
    FNAP = FNA
    FNBP = FNB
    TD[0]=FNAP.real
    TD[1]=FNAP.imag
    TE[0]=FNBP.real
    TE[1]=FNBP.imag
    T[0] = 1.50
        
    TB[0] = T[0] * TB[0]
    TB[1] = T[0] * TB[1]
    TC[0] = T[0] * TC[0]
    TC[1] = T[0] * TC[1]
        
    for  J in range(JX):
        ELTRMX[0,J,0] = TB[0] * PI[1,J] + TC[0] * TAU[1,J]
        ELTRMX[1,J,0] = TB[1] * PI[1,J] + TC[1] * TAU[1,J]
        ELTRMX[2,J,0] = TC[0] * PI[1,J] + TB[0] * TAU[1,J]
        ELTRMX[3,J,0] = TC[1] * PI[1,J] + TB[1] * TAU[1,J]
        ELTRMX[0,J,1] = TB[0] * PI[1,J] - TC[0] * TAU[1,J]
        ELTRMX[1,J,1] = TB[1] * PI[1,J] - TC[1] * TAU[1,J]
        ELTRMX[2,J,1] = TC[0] * PI[1,J] - TB[0] * TAU[1,J]
        ELTRMX[3,J,1] = TC[1] * PI[1,J] - TB[1] * TAU[1,J]


    QEXT   = 2.0 * ( TB[0] + TC[0])
    QSCAT  = ( TB[0]**2 + TB[1]**2 + TC[0]**2 + TC[1]**2 ) / 0.75
    CTBRQS = 0.0
    QBSR   = -2.0*(TC[0] - TB[0])
    QBSI   = -2.0*(TC[1] - TB[1])
    RMM    = -1.0
    N = 2
    while N <= NMX2:
        T[0] = 2*N - 1
        T[1] =   N - 1
        T[2] = 2*N + 1

        for  J in range(JX):
            PI[2,J]  = ( T[0] * PI[1,J] * CSTHT[J] - N * PI[0,J] ) / T[1]
            TAU[2,J] = CSTHT[J] * ( PI[2,J] - PI[0,J] )  - T[0] * SI2THT[J] * PI[1,J]  +  TAU[0,J]
        WM1    =  WFN[0]
        WFN[0] =  WFN[1]
        TA[0]  =  WFN[0].real

        TA[1]  =  WFN[0].imag
        TA[3]  =  WFN[1].imag
        WFN[1] =  T[0] * RX * WFN[0]  -  WM1
        TA[2]  =  WFN[1].real
        
        
        if IFLAG != 2:
            DH2  =  - N / Z[1]  +  1.0 / ( (N / Z[1]) - DH2 )
            DH4  =  - N / Z[3]  +  1.0 / ( (N / Z[3]) - DH4 )
            DH1  =  - N / Z[0]  +  1.0 / ( (N / Z[0]) - DH1 )
            PSTORE  =  ( DH4 + (N / Z[3] ))  *  ( W[2,N-1] + (N / Z[3] ))
            P24H24  =  P24H24 / PSTORE
            HSTORE  =  ( DH1 + (N / Z[0] ))  *  ( W[2,N-1] + (N / Z[3] ))
            P24H21  =  P24H21 / HSTORE
            PSTORE  =  ( ACAP[N-1] + (N / Z[0] ))  /  ( W[2,N-1] + (N / Z[3] ))
            DUMMY   =  DUMMY * PSTORE
            DUMSQ   =  DUMMY * DUMMY
            U[0] =  K3 * ACAP[N-1]  -  K2 * W[0,N-1]
            U[1] =  K3 * ACAP[N-1]  -  K2 * DH2
            U[2] =  K2 * ACAP[N-1]  -  K3 * W[0,N-1]
            U[3] =  K2 * ACAP[N-1]  -  K3 * DH2
            U[4] =  K1 *  W[2,N-1]  -  K2 * W[1,N-1]
            U[5] =  K2 *  W[2,N-1]  -  K1 * W[1,N-1]
            U[6] =  complex( 0.0,-1.0 )  *  ( DUMMY * P24H21 - P24H24 )
            U[7] =  TA[2] / WFN[1]

            FNA  =  U[7] * ( U[0]*U[4]*U[6]  +  K1*U[0]  -  DUMSQ*K3*U[4] ) /( U[1]*U[4]*U[6]  +  K1*U[1]  -  DUMSQ*K3*U[4] )
            FNB  =  U[7] * ( U[2]*U[5]*U[6]  +  K2*U[2]  -  DUMSQ*K2*U[5] ) /( U[3]*U[5]*U[6]  +  K2*U[3]  -  DUMSQ*K2*U[5] )
            TB[0]=FNA.real
            TB[1]=FNA.imag
            TC[0]=FNB.real
            TC[1]=FNB.imag
        TC1  =  ACAP[N-1] * RRF  +  N * RX
        TC2  =  ACAP[N-1] * RF   +  N * RX
        FN1  =  ( TC1 * TA[2]  -  TA[0] ) /  ( TC1 * WFN[1] - WFN[0] )
        FN2  =  ( TC2 * TA[2]  -  TA[0] ) /  ( TC2 * WFN[1] - WFN[0] )
        M    =  int(WVNO * R)
        if  N >= M : 
            if IFLAG ==2: 
                FNA  =  FN1
                FNB  =  FN2
                TB[0]=FNA.real
                TB[1]=FNA.imag
                TC[0]=FNB.real
                TC[1]=FNB.imag

            if IFLAG != 2:    
                if  abs(  ( FN1-FNA ) / FN1  ) < EPSILON_MIE:
                    if abs(  ( FN2-FNB ) / FN2  )  < EPSILON_MIE :
                        IFLAG = 2
                    
            
     
        T[4]  =  N
        T[3]  =  T[0] / ( T[4] * T[1] )
        T[1]  =  (  T[1] * ( T[4] + 1.0 )  ) / T[4]

        CTBRQS +=  T[1] * ( TD[0] * TB[0]  +  TD[1] * TB[1] + TE[0] * TC[0]  +  TE[1]* TC[1] )+T[3] * ( TD[0] * TE[0]  +  TD[1] * TE[1] )
        QEXT   +=    T[2] * ( TB[0] + TC[0] )
        T[3]    =  TB[0]**2 + TB[1]**2 + TC[0]**2 + TC[1]**2
        QSCAT  +=  T[2] * T[3]
        RMM     =  -RMM
        QBSR +=  T[2]*RMM*(TC[0] - TB[0])
        QBSI  +=  T[2]*RMM*(TC[1] - TB[1])

        T[1]    =  N * (N+1)
        T[0]    =  T[2] / T[1]
        K=int(N)
        for J in range(JX):
            ELTRMX[0,J,0] += T[0]*(TB[0]*PI[2,J]+TC[0]*TAU[2,J])
            ELTRMX[1,J,0] += T[0]*(TB[1]*PI[2,J]+TC[1]*TAU[2,J])
            ELTRMX[2,J,0] += T[0]*(TC[0]*PI[2,J]+TB[0]*TAU[2,J])
            ELTRMX[3,J,0] += T[0]*(TC[1]*PI[2,J]+TB[1]*TAU[2,J])
            if  K%2 == 0:
                ELTRMX[0,J,1] += T[0]*(-TB[0]*PI[2,J]+TC[0]*TAU[2,J])
                ELTRMX[1,J,1] += T[0]*(-TB[1]*PI[2,J]+TC[1]*TAU[2,J])
                ELTRMX[2,J,1] += T[0]*(-TC[0]*PI[2,J]+TB[0]*TAU[2,J])
                ELTRMX[3,J,1] += T[0]*(-TC[1]*PI[2,J]+TB[1]*TAU[2,J])
            else:
                ELTRMX[0,J,1] += T[0]*(TB[0]*PI[2,J]-TC[0]*TAU[2,J])
                ELTRMX[1,J,1] += T[0]*(TB[1]*PI[2,J]-TC[1]*TAU[2,J])
                ELTRMX[2,J,1] += T[0]*(TC[0]*PI[2,J]-TB[0]*TAU[2,J])
                ELTRMX[3,J,1] += T[0]*(TC[1]*PI[2,J]-TB[1]*TAU[2,J])
        
        if  T[3] >= EPSILON_MIE:
            
            N += 1
            for  J in range(JX):
                PI[0,J]   =   PI[1,J]
                PI[1,J]   =   PI[2,J]
                TAU[0,J]  =  TAU[1,J]
                TAU[1,J]  =  TAU[2,J]
             
   
            FNAP  =  FNA
            FNBP  =  FNB
            TD[0]=FNAP.real
            TD[1]=FNAP.imag
            TE[0]=FNBP.real
            TE[1]=FNBP.imag
        
        else:
            break 
    if  N >= NMX2 :
        #print(T[3],NMX2)
        istatus=-1
        return 0,0,0,istatus   
    for J in range(JX):
        for K in range(2):
            for I in range(4):
                T[I]  =  ELTRMX[I,J,K]
    
            ELTRMX[1,J,K]  =      T[0]**2  +  T[1]**2
            ELTRMX[0,J,K]  =      T[2]**2  +  T[3]**2
            ELTRMX[2,J,K]  =  T[0] * T[2]  +  T[1] * T[3]
            ELTRMX[3,J,K]  =  T[1] * T[2]  -  T[3] * T[0]
  
    T[0]    =    2.0 * RX**2
    QEXT    =   QEXT * T[0]
    QSCAT   =  QSCAT * T[0]
    CTBRQS  =  2.0 * CTBRQS * T[0]           
    istatus = 0
    
    return QEXT,QSCAT,CTBRQS,istatus


def calc_new_mieff(wave_in, nn,kk, radius, bin_min, bin_max, fort_calc_mie = False):
    #print('binning is back')
    ## Calculates optics by reading refrind files
    thetd=0.0   # incident wave angle
    n_thetd=1
    #number of radii sub bins in order to smooth out fringe effects 
    sub_radii = 6
    
    nradii = len(radius)
    nwave=len(wave_in)  #number of wavalength bin centres for calculation
    
    qext = np.zeros((nwave,nradii))
    qscat = np.zeros((nwave,nradii))
    cos_qscat = np.zeros((nwave,nradii))
        
    #compute individual parameters for each gas
    for iwave in range(nwave):
        for irad in range(nradii):

            #create the 6 sub bins 
            dr5 = (bin_max[irad]-bin_min[irad])/(sub_radii-1) # calculate the spacing in between sub-bins. These will be evenly spread about the mean radius of each bin.        
            rr= bin_min[irad] # initial particle radius to test (smallest value in bin, in cm)

            corerad = 0.
            corereal = 1.
            coreimag = 0.
        ## averaging over 6 radial bins to avoid fluctuations
            if fort_calc_mie:
                wave=wave_in*1e-4  ## converting to cm 
                bad_mie= False
                for isub in range(sub_radii):
                            wvno=2*np.pi/wave[iwave]
                            qe_pass, qs_pass, c_qs_pass,istatus= fort_mie_calc(rr, nn[iwave], kk[iwave], thetd, n_thetd, corerad, corereal, coreimag, wvno)
                            if istatus == 0:
                                    qe=qe_pass
                                    qs=qs_pass
                                    c_qs=c_qs_pass
                            else:
                                if bad_mie == False :
                                    bad_mie = True
                                    print ('do_optics(): no Mie solution. So previous grid value assigned')
                                    ## The mie_calc routine fails to converge if the real refractive index is smaller than 1. This is true the
                                    ## fortran counterpart as well. So previous step values are assigned

                            qext[iwave,irad]+= qe
                            qscat[iwave,irad]+= qs
                            cos_qscat[iwave,irad] += c_qs
                            rr+=dr5
            #this is the default.
            #if no fortran crappy code, use PyMieScatt which does a much faster 
            #more robust computation of the Mie parameters
            else:
                wave=wave_in*1e3  ## converting to nm
                ## averaging over 6 radial bins to avoid fluctuations
                for isub in range(sub_radii):
                    #arr = qext, qsca, qback, g,
                    m_eff = nn[iwave]-(1j)*kk[iwave]  # miepython uses negative k convention
                    x_fac = 2 * np.pi * rr*1e7 / wave[iwave]  # both wave and rr are in nm here
                    arr = mie.efficiencies_mx(m_eff, x_fac)

                    qext[iwave,irad]+= arr[0]
                    qscat[iwave,irad]+= arr[1]
                    cos_qscat[iwave,irad] += arr[3]*arr[1] 
                    rr+=dr5

            ## adding to master arrays
            qext[iwave,irad] = qext[iwave,irad] / sub_radii     
            qscat[iwave,irad] = qscat[iwave,irad] / sub_radii
            cos_qscat[iwave,irad] = cos_qscat[iwave,irad]/ sub_radii

    return qext, qscat, cos_qscat

def calc_new_mieff_optool(wave,radius,bin_min,bin_max,gas,optool_dir,aggregates=False,Df=None,N_mon=None,r_mon=None,k0=0):
    """
    Parameters
    ----------
    wave : array
        wavelength grid, in microns
    radius : array
        effective (mean) particle radius, in centimeters 
            -- if aggregates = False, this is the spherical particle radius
            -- if aggregates = True, this is the radius of a sphere of equivalent volume to the fractal aggregate
    bin_min : array
        minimum radius in each particle bin
    bin_max : array
        maximum radius in each particle bin
    gas : string
        particle species to condense
    optool_dir : string
        directory location of compiled OPTOOL code. Assumes that .lnk files (the optool equivalent to .refrind) are stored in
                a subdirectory within this folder.
                -- NOTE: these .lnk files must be in ascending wavelength order to correctly compute Mie coefficients! 
                -- However, we will flip the arrays once calculated, so that .mieff files are consistently saved in order of descending wavelength.
    Df : float
        the fractal dimension, if aggregrates = True
    N_mon : int
        the number of monomers. If None, must set r_mon directly.
    r_mon : float
        monomer particle radius (cm), used if aggregates = True. Can set directly or calculate from N_mon.
    k0 : float
        the fractal prefactor, either prescribed by user or calculated using Tazaki (2021) Eq 2 

    Returns
    -------
    qext: array
        extinction effiencies for all particle radii/wavelengths
    qsca: array
        scattering effiencies for all particle radii/wavelengths
    cos_qscat: array
        average asymmetry parameter x Q_sca for all particle radii/wavelengths
    """
    nwave=len(wave)
    nradii = len(radius)
    sub_radii = 6

    #set up the empty arrays
    qext = np.zeros((nwave,nradii))#sub_radii
    qscat = np.zeros((nwave,nradii))
    cos_qscat = np.zeros((nwave,nradii))

    # determine whether or not we need to calculate N_mon for each radius
    if N_mon is not None:
        N_mon_prescribed=1 # N_mon is a fixed value, prescribed by the user
        original_N_mon = N_mon # record the original set value, in case we need to reduce it temporarily for the smallest particles in the grid
    else:
        N_mon_prescribed=0 # N_mon needs to be calculated for each new radius
        original_N_mon = 0 # original number of monomers not set

    # determine whether or not we need to calculate k0 for each radius
    if k0>0:
        k_0_prescribed=1 # k0 is prescribed by the user
    else:
        k_0_prescribed=0 # k0 needs to be calculated for each new radius

    # calculate the largest monomer radius required (so we can find the largest size parameter below!)
    if (N_mon_prescribed==1):
        r_mon_max = np.max(radius) / np.cbrt(N_mon) # if we have provided N_mon, calculate r_mon from the largest value in the array(through conservation of volume between an aggregate and a compact sphere)
    else: #user prescribed r_mon
        r_mon_max = r_mon

    # determine largest size parameter required
    min_wavelength = pd.read_csv(f"{optool_dir}/lnk_data/{gas}_VIRGA.lnk", skiprows=1, header=None, sep='\s+', usecols=[0]).values.min() # find minimum wavelength from the lnk data file
    largest_size_parameter_x = 2.0*np.pi*r_mon_max*1e4/min_wavelength # find largest size parameter: remember to convert largest monomer radius from cm to um
    
    if (aggregates==True) and (largest_size_parameter_x > 200):
        print(f'\n\n\nWARNING!!!! The largest radius is {radius[-1]*10000.0:.3e} um, leading to a monomer radius of {r_mon_max*10000.0:.3e} um, and the smallest wavelength')
        print(f'in the lnk data file is {min_wavelength:.3e} um. This means that the MONOMER size parameter x_mon will be {largest_size_parameter_x:.3f}. Optool will struggle')
        print(f'with memory to calculate anything where x_mon > 200, but this is in the geometric limit, where Q_ext values are constant. We')
        print('will therefore assume constant values above x>200, but because x_mon is wavelength-dependent we have to send all data to optool')
        print('one wavelength at a time, which prevents optool using multiple cores and can slow it down signficantly. To prevent this decrease')
        print(f'the max radius or remove some of the lowestwavelengths from your .refrind files (especially if, for example, you have UV/visible')
        print("wavelengths in there from a big experimental dataset but aren't using them)!\n\n")

        print(f'\n Mean radius (um)  Sub-bin radius (um)   N_mon    Fractal dimension    Radius of gyration (um)    k0        r_mon (um)')



        for irad in range(nradii):

            radius_print_switch=0 # reset switch to control which statements are printed to user

            print('----------------------------------------------------------------------------------------------------------------------')
     
            # calculate monomer size in this radius bin (needed to check whether we are in the geometric limit for each wavelength)
            N_mon, r_mon = find_N_mon_and_r_mon(N_mon_prescribed, radius[irad], original_N_mon, r_mon)
            r_mon_micron = r_mon * 1e4 # convert r_mon from cm into um

            # go through each wavelength one at a time
            for iwave in range(nwave): 

                # check whether we are in the geometric limit for this particular wavelength/radius (note that this condition is based on MONOMER radius, not the aggregate compact radius, because that's what the matrix size in MMF depends on)
                size_parameter_monomer = 2*np.pi*r_mon_micron/wave[iwave] # radius and wavelengh are both in um here
                
                # if we have reached the geometric limit (this anticipates and prevents memory overflow)
                if (size_parameter_monomer>200):
                    
                    # check whether a previously calculated value exists
                    if irad>0:
                            
                        # if so, use the last calculated values for every wavelength (and reshape the output arrays in the same format as output by PyMieScatt)
                        qext[iwave,irad] += qext[iwave,irad-1]
                        qscat[iwave,irad] += qscat[iwave,irad-1] # no need to multiply each of these by subradii (e.g. 6 bins) because this is BEFORE the smoothing step at the end (we are still in the wavelength-radius loop)
                        cos_qscat[iwave,irad] += cos_qscat[iwave,irad-1]
                        
                        if(iwave==0): #only print statement for first wavelength
                            print(f'  {10000.0*radius[irad]:10.5f}          GEOMETRIC         REGIME:         USING                 PREVIOUS         SUBRADII      VALUES ')
                            radius_print_switch=1 # set switch to 1 so that we still print the regular statements for the non-geometric regime radii below
                    else:
                        print("\n\n\n WARNING: There are no previously calculated Q_ext values to use, and we are in the geometric limit. Make maximum radius smaller or minimum wavelength larger.")
                        wait = input("Press Enter to continue.")

                else: # for all other cases,run the regular code but for individual wavelengths

                    #create the 6 sub bins 
                    dr5 = (bin_max[irad]-bin_min[irad])/(sub_radii-1) # calculate the spacing in between sub-bins. These will be evenly spread about the mean radius of each bin.
                    dr5_micron = dr5*1e4 # sub-bin radius in um
                    
                    rr= bin_min[irad] # start at minimum of bin (radius in cm)
                    rr_micron = rr * 1e4 # starting radius of bin in um (we will increase this with each iteration of sub-bin loop)

                    #print(f'rr_micron: {rr_micron}')
                    #print(f'dr5_micron: {dr5_micron} um')

                    #loop through the radius sub bins
                    for isub in range(sub_radii):

                        # work out optical parameters at each sub-grid particle size
                        if (aggregates==True):

                            # calculate N_mon and r_mon for this particular sub-radius
                            N_mon, r_mon = find_N_mon_and_r_mon(N_mon_prescribed, rr, original_N_mon, r_mon)
                            r_mon_micron = r_mon * 1e4 # convert r_mon from cm into um

                            # create a list of arguments to pass to optool
                            if(N_mon==1): # SPHERES: single monomer case - basically just doing mie theory here, but via OPTOOL instead of PyMie
                                if(iwave==0) or (radius_print_switch==1): #only print statement for first wavelength or if we were in the geometric regime but are now out of it
                                    print(f'  {10000.0*radius[irad]:10.5f}         {rr_micron:10.5f}    {N_mon:10.0f}           JUST                   DOING            SINGLE       SPHERES ')
                                argument_list = ['-c','lnk_data/'+gas+'_VIRGA.lnk','-a', str(rr_micron),'-l',str(wave[iwave]),'-mie', '-q'] # Mie theory (single spherical particle)
                            elif rr_micron <= r_mon_micron: 
                                N_mon=1 # SPHERES: aggregates cannot be smaller than 1 monomer, so just assume spheres if r < r_mon. Also, set N_mon = 1 (this is not currently used in any further calculations, but is just added here to be clear to the user about what we are doing, as well as being best practice in case we do add code below this part)
                                if(iwave==0) or (radius_print_switch==1): #only print statement for first wavelength or if we were in the geometric regime but are now out of it
                                    print(f'  {10000.0*radius[irad]:10.5f}         {rr_micron:10.5f}    {N_mon:10.0f}           JUST                   DOING            SINGLE       SPHERES ')
                                argument_list = ['-c','lnk_data/'+gas+'_VIRGA.lnk','-a', str(rr_micron),'-l',str(wave[iwave]),'-mie', '-q'] # Mie theory (single spherical particle)

                            else: # use MMF for aggregates

                                # if fractal prefactor k0 is left unprescribed by user, calculate it here
                                if(k_0_prescribed==0):
                                    k0 = calculate_k0(N_mon, Df)
                                
                                # calculate radius of gyration (in um)
                                r_gyro = (N_mon/k0)**(1/Df) * r_mon_micron 

                                if(iwave==0) or (radius_print_switch==1): #only print statement for first wavelength or if we were in the geometric regime but are now out of it
                                    print(f'  {10000.0*radius[irad]:10.5f}         {rr_micron:10.5f}    {N_mon:10.0f}      {Df:10.3f}              {r_gyro:10.5f}      {k0:10.3f}      {r_mon_micron:10.7f} ')
                                argument_list = ['-c','lnk_data/'+gas+'_VIRGA.lnk','-a', str(rr_micron),'-l',str(wave[iwave]),'-mmf',str(r_mon_micron),str(Df),str(k0), '-q'] # MMF theory (aggregates): -c, link to refractive indices, -a, radius, -l, wavelength, -mmf, monomer radius r0, fractal dimension d_f, fractal prefactor k0
                        else:
                                if(iwave==0) or (radius_print_switch==1): #only print statement for first wavelength or if we were in the geometric regime but are now out of it
                                    print(f'  {10000.0*radius[irad]:10.5f}         {rr_micron:10.5f}    {N_mon:10.0f}           JUST                   DOING            SINGLE       SPHERES ')
                                argument_list = ['-c','lnk_data/'+gas+'_VIRGA.lnk','-a', str(rr_micron),'-l',str(wave[iwave]),'-mie', '-q'] # Mie theory (single spherical particle)

                        # run optool using the string created above - this finds kext[0], ksca[0], kabs[0], gsca[0], one particle radius at a time, for all wavelengths in the list, and stores them in a table within file "dustkappa.dat" (written in the optool directory)
                        subprocess.run([f"{optool_dir}/optool"]+argument_list, cwd=optool_dir) # code executable + arguments, complete in the working directory of the compiled OPTOOL code 

                        # read the wavelength, k_abs, k_sca, and g_asymmetry values in the 'dustkappa.dat' file that was just created by optool
                        arr=pd.read_csv(optool_dir+"/dustkappa.dat", names=['lambda','k_abs','k_sca','g_asymmetry'], delimiter=r"\s+", comment='#') # skip the comments in the header of the refractive index file, and read in the data
                        arr= arr.drop([0,1]).reset_index(drop=True) # drop first two rows (data not needed) and re-index

                        # read the density from the header of the material file
                        density=pd.read_csv(optool_dir+'/lnk_data/'+gas+'_VIRGA.lnk',nrows=1, delimiter=r"\s+", comment='#', header=None) # read the line after the comments and before the data in the refractive index file (contains "num wavelengths, density" in that order)
                        rho_p = density[1][0] # store the density of the condensate in g/cm3

                        #convert k_abs, k_sca and g from optool output into Q values. Radius needs to be in centimeters here for this conversion since rho_p has g/cm3 units
                        qe = (4 * (arr["k_abs"]+arr["k_sca"]) * rho_p * rr) / 3
                        qs = (4 * arr["k_sca"] * rho_p * rr) / 3
                        qa = (4 * arr["k_abs"] * rho_p * rr) / 3
                        cos_qs = arr["g_asymmetry"] * qs

                        #increment over the radius sub bins
                        rr += dr5  
                        rr_micron += dr5_micron
                        
                        #reshape the output arrays in the same format as output by PyMieScatt -- this will sum all values for all 6 sub-bins into a single array element, and we find te average of the whole array outside of the radius loop
                        qext[iwave,irad] += float(qe.iloc[0])
                        qscat[iwave,irad] += float(qs.iloc[0])
                        cos_qscat[iwave,irad] += float(cos_qs.iloc[0])
                    
                    radius_print_switch=0 # reset witch to control which statements are printed to user
        
        #smooth over all the radius bins 
        qext = qext / sub_radii  #matrix as a function of [iwave,irad] divide by 6 (the number of subradii bins)
        qscat = qscat / sub_radii  #matrix as a function of [iwave,irad]   
        cos_qscat = cos_qscat / sub_radii  #matrix as a function of [iwave,irad]   

    else: # regular method (no size parameter mitigation needed to prevent memory overflow)

        print(f'\n Mean radius (um)  Sub-bin radius (um)   N_mon    Fractal dimension    Radius of gyration (um)    k0        r_mon (um)')

        for irad in range(nradii):
                    
            print('----------------------------------------------------------------------------------------------------------------------')

            #create the 6 sub bins 
            dr5 = (bin_max[irad]-bin_min[irad])/(sub_radii-1) # calculate the spacing in between sub-bins. These will be evenly spread about the mean radius of each bin.
            dr5_micron = dr5*1e4 # sub-bin radius in um
            
            rr= bin_min[irad] # start at minimum of bin (radius in cm)
            rr_micron = rr * 1e4 # starting radius of bin in um (we will increase this with each iteration of sub-bin loop)

            #print(f'rr_micron: {rr_micron}')
            #print(f'dr5_micron: {dr5_micron} um')

            #loop through the radius sub bins
            for isub in range(sub_radii):

                # work out optical parameters at each sub-grid particle size
                if (aggregates==True):

                    # calculate N_mon and r_mon for this particular sub-radius
                    N_mon, r_mon = find_N_mon_and_r_mon(N_mon_prescribed, rr, original_N_mon, r_mon)
                    r_mon_micron = r_mon * 1e4 # convert r_mon from cm into um

                    # create a list of arguments to pass to optool
                    if(N_mon==1): # SPHERES: single monomer case - basically just doing mie theory here, but via OPTOOL instead of PyMie
                        print(f'  {10000.0*radius[irad]:10.5f}         {rr_micron:10.5f}    {N_mon:10.0f}           JUST                   DOING            SINGLE       SPHERES ')
                        argument_list = ['-c','lnk_data/'+gas+'_VIRGA.lnk','-a', str(rr_micron),'-l','lnk_data/'+gas+'_VIRGA.lnk','-mie', '-q'] # Mie theory (single spherical particle)
                    elif rr_micron <= r_mon_micron: 
                        N_mon=1 # SPHERES: aggregates cannot be smaller than 1 monomer, so just assume spheres if r < r_mon. Also, set N_mon = 1 (this is not currently used in any further calculations, but is just added here to be clear to the user about what we are doing, as well as being best practice in case we do add code below this part)
                        print(f'  {10000.0*radius[irad]:10.5f}         {rr_micron:10.5f}    {N_mon:10.0f}           JUST                   DOING            SINGLE       SPHERES ')
                        argument_list = ['-c','lnk_data/'+gas+'_VIRGA.lnk','-a', str(rr_micron),'-l','lnk_data/'+gas+'_VIRGA.lnk','-mie', '-q'] # Mie theory (single spherical particle)

                    else: # use MMF for aggregates

                        # if fractal prefactor k0 is left unprescribed by user, calculate it here
                        if(k_0_prescribed==0):
                            k0 = calculate_k0(N_mon, Df)
                        
                        # calculate radius of gyration (in um)
                        r_gyro = (N_mon/k0)**(1/Df) * r_mon_micron 

                        print(f'  {10000.0*radius[irad]:10.5f}         {rr_micron:10.5f}    {N_mon:10.0f}      {Df:10.3f}              {r_gyro:10.5f}      {k0:10.3f}      {r_mon_micron:10.7f} ')
                        argument_list = ['-c','lnk_data/'+gas+'_VIRGA.lnk','-a', str(rr_micron),'-l','lnk_data/'+gas+'_VIRGA.lnk','-mmf',str(r_mon_micron),str(Df),str(k0), '-q'] # MMF theory (aggregates): -c, link to refractive indices, -a, radius, -l, list of wavelengths (do whole file), -mmf, monomer radius r0, fractal dimension d_f, fractal prefactor k0
                else:
                        print(f'  {10000.0*radius[irad]:10.5f}         {rr_micron:10.5f}    {N_mon:10.0f}           JUST                   DOING            SINGLE       SPHERES ')
                        argument_list = ['-c','lnk_data/'+gas+'_VIRGA.lnk','-a', str(rr_micron),'-l','lnk_data/'+gas+'_VIRGA.lnk','-mie', '-q'] # Mie theory (single spherical particle)

                # run optool using the string created above - this finds kext[0], ksca[0], kabs[0], gsca[0], one particle radius at a time, for all wavelengths in the list, and stores them in a table within file "dustkappa.dat" (written in the optool directory)
                subprocess.run([f"{optool_dir}/optool"]+argument_list, cwd=optool_dir) # code executable + arguments, complete in the working directory of the compiled OPTOOL code 

                # read the wavelengths, k_abs, k_sca, and g_asymmetry values in the 'dustkappa.dat' file that was just created by optool
                arr=pd.read_csv(optool_dir+"/dustkappa.dat", names=['lambda','k_abs','k_sca','g_asymmetry'], delimiter=r"\s+", comment='#') # skip the comments in the header of the refractive index file, and read in the data
                arr= arr.drop([0,1]).reset_index(drop=True) # drop first two rows (data not needed) and re-index

                #print(arr)

                # read the density from the header of the material file
                density=pd.read_csv(optool_dir+'/lnk_data/'+gas+'_VIRGA.lnk',nrows=1, delimiter=r"\s+", comment='#', header=None) # read the line after the comments and before the data in the refractive index file (contains "num wavelengths, density" in that order)
                rho_p = density[1][0] # store the density of the condensate in g/cm3

                #convert k_abs, k_sca and g from optool output into Q values. Radius needs to be in centimeters here for this conversion since rho_p has g/cm3 units
                qe = (4 * (arr["k_abs"]+arr["k_sca"]) * rho_p * rr) / 3
                qs = (4 * arr["k_sca"] * rho_p * rr) / 3
                qa = (4 * arr["k_abs"] * rho_p * rr) / 3
                cos_qs = arr["g_asymmetry"] * qs

                #increment over the radius sub bins
                rr += dr5  
                rr_micron += dr5_micron
                
                #reshape the output arrays in the same format as output by PyMieScatt
                for iwave in range(nwave):
                    qext[iwave,irad] += qe[iwave]
                    qscat[iwave,irad] += qs[iwave]
                    cos_qscat[iwave,irad] += cos_qs[iwave]

        
        #smooth over all the radius bins 
        qext = qext / sub_radii  #matrix as a function of [iwave,irad] divide by 6 (the number of subradii bins)
        qscat = qscat / sub_radii  #matrix as a function of [iwave,irad]   
        cos_qscat = cos_qscat / sub_radii  #matrix as a function of [iwave,irad]   
                
    return qext, qscat, cos_qscat
